import logging

from crownstone_core.Exceptions import CrownstoneException
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.Conversion import Conversion

from crownstone_uart.core.uart.UartTypes import UartTxType

_LOGGER = logging.getLogger(__name__)

class UartMessagePacket:
	"""
	UART message packet:
	2B data type (aka opcode)
	xB data
	"""

	def __init__(self, opCode: UartTxType = UartTxType.UNKNOWN, payload: list or bytearray = None):
		"""
		:param opCode:       uint16 data type.
		:param payload:      List of uint8, according to data type
		"""
		self.opCode = opCode
		if payload is None:
			self.payload = []
		else:
			self.payload = payload

	def parse(self, data: list):
		"""
		Parses data.

		:returns True on success.
		"""
		try:
			reader = BufferReader(data)
			self.opCode = reader.getUInt16()
			self.payload = reader.getRemainingBytes()
			return True
		except CrownstoneException as e:
			_LOGGER.warning(F"Parse error: {e}")
			return False

	def serialize(self):
		# Header: 1B device ID, 2B opcode

		# construct the basePacket, which is used for CRC calculation
		uartMsg = []
		uartMsg += Conversion.uint16_to_uint8_array(self.opCode)
		uartMsg += self.payload

		return uartMsg



class UartTopics:

    newDataAvailable = "UART_newDataAvailable"

    uartMessage = "UART_Message" # data is dictionary: {"string": str, "data": [uint8, uint8, ...] }

    hello = "UART_hello" # Data is: UartCrownstoneHelloPacket

    log = "UART_log" # Data is UartLogPacket
    logArray = "UART_logArray" # Data is UartLogArrayPacket

    assetIdReport = "assetIdReport" # Data is an AssetIdReport class instance.
    assetTrackingReport = "assetTrackingReport" # Data is an AssetMacReport class instance.

    rssiBetweenStones = "rssiBetweenStones" # Data is RssiBetweenStonesPacket

    microappMessage = "Microapp_message" # Data is MicroappMessagePacket



#!/usr/bin/env python3

"""An example that switches a Crownstone, and prints the power usage of the selected Crownstone."""

import time

from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType

from crownstone_uart import CrownstoneUart, UartEventBus, UartTopics


# This is the id of the Crownstone we will be switching
# change it to match the Crownstone Id you want to switch!
targetCrownstoneId = 3

def showNewData(data):
	global targetCrownstoneId
	if data.crownstoneId == targetCrownstoneId:
		print("New data received!")
		if data.type == AdvType.CROWNSTONE_STATE or data.type == AdvType.EXTERNAL_STATE:
			print(f"PowerUsage of crownstone {data.crownstoneId} is {data.powerUsageReal}W")
		print("-------------------")


uart = CrownstoneUart()

# Start up the USB bridge.
uart.initialize_usb_sync()
# you can alternatively do this async by
# await uart.initialize_usb()

# Set up event listeners
UartEventBus.subscribe(UartTopics.newDataAvailable, showNewData)

# Switch this Crownstone on and off.
turnOn = True

# The try except part is just to catch a control+c, time.sleep does not appreciate being killed.
try:
	for i in range(0, 10):
		if not uart.running:
			break

		if turnOn:
			print("Switching Crownstone on  (iteration: ", i,")")
		else:
			print("Switching Crownstone off (iteration: ", i,")")
		uart.switch_crownstone(targetCrownstoneId, on = turnOn)

		turnOn = not turnOn
		time.sleep(2)
except KeyboardInterrupt:
	print("\nClosing example.... Thanks for your time!")

uart.stop()



from crownstone_core.util.Conversion import Conversion

COUNTER_FREQUENCY = 1.0 / 32768.0

class CurrentSamplesPacket:

    amountOfSamples = 100
    sampleSize = 2
    timestampSize = 4
    packetSize = amountOfSamples * sampleSize + timestampSize

    typeDescription = 'current'


    def __init__(self, payload):
        if len(payload) < self.packetSize:
            print("ERROR: INVALID PAYLOAD LENGTH", len(payload), payload)
            return
        
        self.timestampCounter = Conversion.uint8_array_to_uint32(payload[0:4])
        self.samples = []
        for i in range(4, self.packetSize, self.sampleSize):
            self.samples.append(Conversion.uint8_array_to_int16(payload[i:i+self.sampleSize]))

    def getDict(self):
        data = {}

        data["id"] = 0 # TODO: get the Crownstone ID here.
        data["type"] = self.typeDescription
        data["timestamp"] = self.timestampCounter
        data["data"] = self.samples

        return data



class SystemTopics:
    newCrownstoneFound    = "newCrownstoneFound"
    stateUpdate           = "stateUpdate"            # used to propagate verified state messages through the system
    
    uartRawData           = 'uartRawData'            # Sent when new raw uart data is received. Data is a byte array.
    uartDiscardedData     = 'uartDiscardedData'      # Sent when uart data is received that was discarded because it's not a valid packet. Data is a byte array.
    uartNewPackage        = 'uartNewPackage'         # Sent when a UART packet is received. Data is a UartWrapperPacket.
    uartNewMessage        = 'uartNewMessage'         # Sent when a UART message is received. Data is a UartMessagePacket.

    uartWriteData         = 'uartWriteData'          # used to write to the UART. Data is array of bytes.

    uartWriteError        = 'uartWriteError'         # used to write to the UART. Data is array of bytes.
    uartWriteSuccess      = 'uartWriteSuccess'       # used to write to the UART. Data is array of bytes.

    resultPacket          = 'resultPacket'           # data is a ResultPacket class instance
    meshResultPacket      = 'meshResultPacket'       # data is a list [Crownstone ID, ResultPacket]
    meshResultFinalPacket = 'meshResultFinalPacket'  # data is a ResultPacket class instance

    connectionClosed      = 'connectionClosed'
    connectionEstablished = 'connectionEstablished'




import asyncio
import logging
from typing import List

from crownstone_core.Exceptions import CrownstoneException, CrownstoneError
from crownstone_core.packets.ResultPacket import ResultPacket
from crownstone_core.packets.assetFilter.FilterCommandPackets import FilterSummaryPacket, FilterSummariesPacket
from crownstone_core.packets.assetFilter.builders.AssetFilter import AssetFilter
from crownstone_core.packets.assetFilter.util import AssetFilterMasterCrc
from crownstone_core.packets.assetFilter.util.AssetFilterChunker import FilterChunker
from crownstone_core.packets.assetFilter.util.AssetFilterSyncer import AssetFilterSyncer

from crownstone_core.protocol.BluenetTypes import ResultValue
from crownstone_core.protocol.ControlPackets import ControlPacketsGenerator
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.dataFlowManagers.Collector import Collector
from crownstone_uart.core.uart.UartTypes import UartTxType, UartMessageType
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket
from crownstone_uart.topics.SystemTopics import SystemTopics

_LOGGER = logging.getLogger(__name__)

class ControlHandler:

    async def factoryReset(self):
        _LOGGER.info("factory reset")
        await self._writeControlAndWaitForSuccess(ControlPacketsGenerator.getCommandFactoryResetPacket())

    async def setFilters(self, filters: List[AssetFilter], masterVersion: int = None) -> int:
        """
        Makes sure the given filters are set at the Crownstone.
        Uploads and removes filters where necessary.
        :param filters:           The asset filter to be uploaded.
        :param masterVersion:     The new master version. If None, the master version will be increased by 1.
        :return:                  The new master version.
        """
        _LOGGER.info(f"setFilters")
        _LOGGER.warning("This function will move to the assetFilter.")
        summaries = await self.getFilterSummaries()
        syncer = AssetFilterSyncer(summaries, filters, masterVersion)
        if not syncer.commitRequired:
            return syncer.masterVersion

        for filterId in syncer.removeIds:
            await self.removeFilter(filterId)

        for filter in filters:
            if filter.getFilterId() in syncer.uploadIds:
                await self.uploadFilter(filter)

        await self.commitFilterChanges(syncer.masterVersion, filters)
        return syncer.masterVersion

    async def getFilterSummaries(self) -> FilterSummariesPacket:
        """
        Get a summary of the filters that are on the Crownstones.
        This can be used to determine:
        - Which filters should be changed.
        - What the next master version should be.
        - How much space there is left for new filters.
        - The new master CRC.

        :return:   The filter summaries packet.
        """
        _LOGGER.info(f"getFilterSummaries")
        _LOGGER.warning("This function will move to the assetFilter.")
        result = await self._write(ControlPacketsGenerator.getGetFilterSummariesPacket())
        if result is None:
            raise CrownstoneException(CrownstoneError.DATA_MISSING, "No summaries received")
        summaries = FilterSummariesPacket(result)
        return summaries

    async def uploadFilter(self, filter: AssetFilter):
        """
        Upload an asset filter to the Crownstones.
        Once all changes are made, don't forget to commit them.

        :param filter:  The asset filter to be uploaded.
        """
        _LOGGER.info(f"uploadFilter {filter}")
        _LOGGER.warning("This function will move to the assetFilter.")
        chunker = FilterChunker(filter, 128)
        result = None
        for i in range(0, chunker.getAmountOfChunks()):
            chunk = chunker.getChunk()
            result = await self._write(ControlPacketsGenerator.getUploadFilterPacket(chunk))
        return result

    async def removeFilter(self, filterId):
        """
        Remove an asset filter from the Crownstones.
        Once all changes are made, don't forget to commit them.

        :param filterId:     The filter ID to be removed.
        """
        _LOGGER.info(f"removeFilter id={filterId}")
        _LOGGER.warning("This function will move to the assetFilter.")
        return await self._write(ControlPacketsGenerator.getRemoveFilterPacket(filterId))

    async def commitFilterChanges(self, masterVersion: int, filters: List[AssetFilter], filterSummaries: List[FilterSummaryPacket] = None):
        """
        Commit the changes made by upload and/or remove.

        :param masterVersion:     The new master version, should be higher than previous master version.
        :param filters:           A list of asset filters with filter ID, that are uploaded to the Crowstone.
        :param filterSummaries :  A list of filter summaries that are already on the Crownstone.
        """
        _LOGGER.info(f"commitFilterChanges masterVersion={masterVersion}")
        _LOGGER.warning("This function will move to the assetFilter.")
        masterCrc = AssetFilterMasterCrc.get_master_crc_from_filters(filters, filterSummaries)
        return await self._write(ControlPacketsGenerator.getCommitFilterChangesPacket(masterVersion, masterCrc))


    async def _write(self, controlPacket: [int], successCodes = [ResultValue.SUCCESS, ResultValue.SUCCESS_NO_CHANGE, ResultValue.WAIT_FOR_SUCCESS]) -> [int] or None:
        """
        Returns the result payload.
        TODO: return result packet.
        TODO: use a ControlPacket as param, instead of int array.
        """
        _LOGGER.debug(f"Write control packet {controlPacket}")
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        resultCollector = Collector(timeout=1, topic=SystemTopics.resultPacket)
        # send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        # wait for the collectors to fill
        commandResultData = await resultCollector.receive()

        if commandResultData is not None:
            if commandResultData.resultCode not in successCodes:
                raise CrownstoneException(commandResultData.resultCode, f"Command has failed: result code is {commandResultData.resultCode}")
            return commandResultData.payload
        return None

    async def _writeControlAndGetResult(
            self,
            controlPacket,
            acceptedResultValues = [ResultValue.SUCCESS, ResultValue.SUCCESS_NO_CHANGE],
            timeout = None
    ) -> ResultPacket:
        """
        Writes the control packet, checks the result value, and returns the result packet.
        :param controlPacket:          Serialized control packet to write.
        :param acceptedResultValues:   List of result values that are ok.
        :param timeout:                Timeout in seconds.
        :returns:                      The result packet.
        """
        _LOGGER.debug(f"Write control packet {controlPacket}")
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        if timeout is None:
            timeout = 3

        # Setup the result collector
        resultCollector = Collector(timeout=timeout, topic=SystemTopics.resultPacket)

        # Send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        # Wait for the result
        resultPacket: ResultPacket = await resultCollector.receive()
        _LOGGER.debug(f"resultPacket={resultPacket}")

        if resultPacket is None:
            raise CrownstoneException(CrownstoneError.TIMEOUT, f"No result received within {timeout} seconds")
        if not resultPacket.valid:
            raise CrownstoneException(CrownstoneError.INCORRECT_RESPONSE_LENGTH, "Result is invalid")
        if resultPacket.resultCode not in acceptedResultValues:
            raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, f"Result code is {resultPacket.resultCode}")
        return resultPacket

    async def _writeControlAndWaitForSuccess(
            self,
            controlPacket,
            timeout = 10,
            acceptedResultValues = [ResultValue.SUCCESS, ResultValue.SUCCESS_NO_CHANGE]
    ) -> ResultPacket:
        """
        Writes the control packet, and waits for success.
        :param controlPacket:          Serialized control packet to write.
        :param timeout:                Timeout in seconds.
        :param acceptedResultValues:   List of result values that are considered a success.
        :returns:                      The result packet.
        """
        _LOGGER.debug(f"Write control packet {controlPacket}")
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        # Setup the result collector
        resultCollector = Collector(timeout=timeout, topic=SystemTopics.resultPacket)

        # Send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        while True:
            # Wait for the result
            resultPacket: ResultPacket = await resultCollector.receiveNext()
            _LOGGER.debug(f"resultPacket={resultPacket}")

            if resultPacket is None:
                resultCollector._cleanup()
                raise CrownstoneException(CrownstoneError.TIMEOUT, f"No success received within {timeout} seconds")
            if resultPacket.resultCode in acceptedResultValues:
                resultCollector._cleanup()
                return resultPacket
            if not resultPacket.valid:
                resultCollector._cleanup()
                raise CrownstoneException(CrownstoneError.INCORRECT_RESPONSE_LENGTH, "Result is invalid")
            if resultPacket.resultCode == ResultValue.WAIT_FOR_SUCCESS:
                continue
            if resultPacket.resultCode not in acceptedResultValues:
                resultCollector._cleanup()
                raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS,
                                          f"Result code is {resultPacket.resultCode}")
            resultCollector._cleanup()
            return resultPacket



import logging
import sys
import threading
import traceback

import serial
import serial.tools.list_ports

from crownstone_uart.Constants import UART_READ_TIMEOUT, UART_WRITE_TIMEOUT
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.UartParser import UartParser
from crownstone_uart.core.uart.UartReadBuffer import UartReadBuffer
from crownstone_uart.topics.SystemTopics import SystemTopics
from crownstone_uart.Exceptions import UartBridgeError, UartException

_LOGGER = logging.getLogger(__name__)


class UartBridge(threading.Thread):
    """
    Thread that manages a serial port instance.
    """

    def __init__(self, exception_queue, port, baudrate, writeChunkMaxSize=0):
        self.bridge_exception_queue = exception_queue
        self.baudrate = baudrate
        self.port = port
        self.writeChunkMaxSize = writeChunkMaxSize

        self.serialController = None
        self.started = False

        self.running = True
        self.parser = UartParser()
        self.eventId = UartEventBus.subscribe(SystemTopics.uartWriteData, self.write_to_uart)

        threading.Thread.__init__(self)

    def __del__(self):
        self.stop()


    def run(self):
        try:
            self.start_serial()
            self.start_reading()
        except (UartException, BaseException):
            self.bridge_exception_queue.put(sys.exc_info())


    def stop(self):
        self.running = False
        UartEventBus.unsubscribe(self.eventId)
        self.parser.stop()


    def start_serial(self):
        _LOGGER.debug(F"UartBridge: Initializing serial on port {self.port} with baudrate {self.baudrate}")
        try:
            self.serialController = serial.Serial()
            self.serialController.port = self.port
            self.serialController.baudrate = int(self.baudrate)
            self.serialController.timeout = UART_READ_TIMEOUT
            self.serialController._write_timeout = UART_WRITE_TIMEOUT
            self.serialController.open()
        except OSError or serial.SerialException or KeyboardInterrupt as serial_error:
            self.stop()
            _LOGGER.error(serial_error)
            raise UartException(UartBridgeError.CANNOT_OPEN_SERIAL_CONTROLLER, serial_error) from serial_error


    def start_reading(self):
        readBuffer = UartReadBuffer()
        self.started = True
        _LOGGER.debug(F"Read starting on serial port {self.port} {self.running}")
        try:
            while self.running:
                bytesFromSerial = self.serialController.read() # reads single byte (blocks until it is received)
                if bytesFromSerial:
                    # read all extra bytes in read buffer
                    if self.serialController.in_waiting > 0:
                        additionalBytes = self.serialController.read(self.serialController.in_waiting)
                        bytesFromSerial = bytesFromSerial + additionalBytes

                    UartEventBus.emit(SystemTopics.uartRawData, bytesFromSerial)
                    readBuffer.addByteArray(bytesFromSerial)
        except OSError or serial.SerialException:
            _LOGGER.info("Connection to USB Failed. Retrying...")
        except KeyboardInterrupt:
            self.running = False
            _LOGGER.debug("Closing serial connection.")
        except Exception as error:
            self.running = False
            _LOGGER.error(f"Error in handling UART data: {error}")
            traceback.print_exc()

        # close the serial controller
        self.serialController.close()
        self.serialController = None
        # remove the event listener pointing to the old connection
        UartEventBus.unsubscribe(self.eventId)
        self.started = False
        UartEventBus.emit(SystemTopics.connectionClosed, True)

    def write_to_uart(self, data):
        _LOGGER.debug(f"write_to_uart: {data}")
        if self.serialController is not None and self.started:
            try:
                if self.writeChunkMaxSize == 0:
                    self.serialController.write(data)
                else:
                    # writing in chunks solves issues writing to certain JLink chips. A max chunkSize of 64 was found to work well for our case.
                    chunkSize = self.writeChunkMaxSize
                    index = 0
                    while (index*chunkSize) < len(data):
                        chunkedData = data[index*chunkSize:chunkSize*(index+1)]
                        index += 1
                        self.serialController.write(chunkedData)

                UartEventBus.emit(SystemTopics.uartWriteSuccess, data)
            except serial.SerialTimeoutException as e:
                UartEventBus.emit(SystemTopics.uartWriteError, {"message":"Timeout on uart write.", "error": e})
            except serial.SerialException as e:
                UartEventBus.emit(SystemTopics.uartWriteError, {"message":"SerialException occurred during uart write", "error": e})
            except OSError as e:
                UartEventBus.emit(SystemTopics.uartWriteError, {"message":"OSError occurred during uart write.", "error": e})
            except Exception as e:
                UartEventBus.emit(SystemTopics.uartWriteError, {"message": "Unknown Exception during uart write.", "error": e})
            except:
                e = sys.exc_info()[0]
                UartEventBus.emit(SystemTopics.uartWriteError, {"message":"Unknown error during uart write.", "error": e})
        else:
            self.stop()



import logging

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader

from crownstone_uart.core.uart.uartPackets.UartLogHeaderPacket import UartLogHeaderPacket

_LOGGER = logging.getLogger(__name__)

class UartLogArrayPacket(BasePacket):
	"""
	UART log array packet
	"""

	def __init__(self, data = None):
		self.header = UartLogHeaderPacket()
		self.elementType = None
		self.elementSize = 0
		self.elementData = None
		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		self.header.deserialize(reader)
		self.elementType = reader.getUInt8()
		self.elementSize = reader.getUInt8()
		self.elementData = reader.getRemainingBytes()

	def __str__(self):
		return f"UartLogArrayPacket(" \
		       f"header={self.header}, " \
		       f"elementType={self.elementType}, " \
		       f"elementSize={self.elementSize}, " \
		       f"elementData={self.elementData})"



from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.topics.SystemTopics import SystemTopics
from crownstone_uart.topics.UartTopics import UartTopics


class StoneStateManager:
    def __init__(self):
        self.stones = {}
        UartEventBus.subscribe(SystemTopics.stateUpdate, self.handleStateUpdate)

    def handleStateUpdate(self, data):
        stoneId    = data[0]
        advPayload = data[1]

        if stoneId in self.stones:
            if hasattr(advPayload, 'timestamp'):
                if self.stones[stoneId].timestamp < advPayload.timestamp:
                    self.stones[stoneId] = advPayload
                    self.emitNewData(advPayload)
        else:
            UartEventBus.emit(SystemTopics.newCrownstoneFound, stoneId)
            self.stones[stoneId] = advPayload
            self.emitNewData(advPayload)
    
    def emitNewData(self, advPayload):
        UartEventBus.emit(UartTopics.newDataAvailable, advPayload)

    def getIds(self):
        ids = []
        for stoneId, stoneData in self.stones.items():
            ids.append(stoneId)

        return ids



from crownstone_core.util.BufferReader import BufferReader

from crownstone_uart.core.uart.uartPackets.UartCrownstoneStatusPacket import UartCrownstoneStatusPacket

class UartCrownstoneHelloPacket:
    """
    UART crownstone hello packet:
    1B sphereId
    1B status
    """
    def __init__(self, data):
        if isinstance(data, BufferReader):
            reader = data
        else:
            reader = BufferReader(data)

        # Parse buffer
        self.sphereId = reader.getUInt8()

        self.status = UartCrownstoneStatusPacket(reader)



from crownstone_core.util.Conversion import Conversion
import logging

from crownstone_core.util.CRC import crc16ccitt
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import SIZE_HEADER_SIZE, CRC_SIZE, WRAPPER_HEADER_SIZE, START_TOKEN, \
    ESCAPE_TOKEN, BIT_FLIP_MASK, UartWrapperPacket
from crownstone_uart.topics.DevTopics import DevTopics
from crownstone_uart.topics.SystemTopics import SystemTopics
from crownstone_uart.util.UartUtil import UartUtil

_LOGGER = logging.getLogger(__name__)


class UartReadBuffer:
    """
    Handles incoming uart data through the add() and addByteArray() methods.
    Start bytes, escaping, packet framing and crc checking are parsed.

    Emitted events:
     - SystemTopics.uartNewPackage: successfully parsed packet.
     - SystemTopics.uartDiscardedData
    """

    def __init__(self):
        # Buffer with all the received bytes since last emit
        self.rawBuffer = []

        # Buffer with the parsed data: excludes start token, bytes are escaped,
        self.buffer = []

        self.escapingNextByte = False

        # Set to True when the start token has been received.
        self.active = False

        # Set when the packet size has been received.
        self.sizeToRead = 0

    def addByteArray(self, rawByteArray):
        for byte in rawByteArray:
            self.add(byte)

    def add(self, byte):
        self.rawBuffer.append(byte)

        # An escape shouldn't be followed by a special byte.
        if self.escapingNextByte and (byte is START_TOKEN or byte is ESCAPE_TOKEN):
            _LOGGER.warning("Received special byte after escape token.")
            self.reset(True)
            return

        # Activate on start token.
        if byte is START_TOKEN:
            if self.active:
                _LOGGER.warning("Received multiple start tokens.")
                self.reset(True)
            else:
                # Reset to be sure.
                self.reset()

            self.active = True
            return

        if not self.active:
            # Since the start token was not received, we discard this byte.
            self.reset(True)
            return

        # Escape next byte on escape token.
        if byte is ESCAPE_TOKEN:
            self.escapingNextByte = True
            return

        if self.escapingNextByte:
            byte ^= BIT_FLIP_MASK
            self.escapingNextByte = False

        self.buffer.append(byte)
        bufferSize = len(self.buffer)

        if self.sizeToRead == 0:
            # We didn't parse the size yet.
            if bufferSize == SIZE_HEADER_SIZE:
                # Now we know the remaining size to read
                self.sizeToRead = Conversion.uint8_array_to_uint16(self.buffer)

                # Size to read shouldn't be 0.
                if self.sizeToRead == 0:
                    self.reset(True)
                    return

                self.buffer = []
                return

        elif bufferSize >= self.sizeToRead:
            processSuccessful = self.process()
            self.reset(not processSuccessful)
            return

    def process(self):
        """
        Process a buffer.

        Check CRC, and emit a uart packet.

        Buffer starts after size header, and includes wrapper header and tail (CRC).

        Return: processing success
        """

        # Check size
        bufferSize = len(self.buffer)
        wrapperSize = WRAPPER_HEADER_SIZE + CRC_SIZE
        if bufferSize < wrapperSize:
            _LOGGER.warning("Buffer too small")
            return False

        # Get the buffer between size field and CRC:
        baseBuffer = self.buffer[0 : bufferSize - CRC_SIZE]

        # Check CRC
        calculatedCrc = crc16ccitt(baseBuffer)
        sourceCrc = Conversion.uint8_array_to_uint16(self.buffer[bufferSize - CRC_SIZE : ])

        if calculatedCrc != sourceCrc:
            _LOGGER.warning("Failed CRC: {0} != {1} (data: {2})".format(calculatedCrc, sourceCrc, baseBuffer))
            return False

        wrapperPacket = UartWrapperPacket()
        if wrapperPacket.parse(baseBuffer):
            UartEventBus.emit(SystemTopics.uartNewPackage, wrapperPacket)
            return True

    def reset(self, discarded=False):
        """
        Resets to initial state, emitting uartDiscardedData if necessary.

        :param discarded: True when reset due to discarded data.
        """
        if discarded and len(self.rawBuffer):
            UartEventBus.emit(SystemTopics.uartDiscardedData, self.rawBuffer)

        self.rawBuffer = []
        self.buffer = []
        self.escapingNextByte = False
        self.active = False
        self.sizeToRead = 0




#!/usr/bin/env python3

"""
An example that checks if the communication with the Crownstone works, by commanding it to send us a message.
"""
import time

# Create new instance of uart
from crownstone_uart import CrownstoneUart, UartEventBus, UartTopics

uart = CrownstoneUart()

# Start up the USB bridge.
uart.initialize_usb_sync()
# you can alternatively do this async by
# await uart.initialize_usb()

# Function that's called when the power usage is updated.
def showUartMessage(data):
	print("Received payload", data)

# Set up event listeners
UartEventBus.subscribe(UartTopics.uartMessage, showUartMessage)

# the try except part is just to catch a control+c, time.sleep does not appreciate being killed.
try:
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
	uart.uart_echo("HelloWorld")
	time.sleep(0.2)
except KeyboardInterrupt:
	print("Closing example.... Thanks for your time!")

uart.stop()


from enum import IntEnum

class UartMessageType(IntEnum):
    UART_MESSAGE =                     0
    ENCRYPTED_UART_MESSAGE =           128

class UartTxType(IntEnum):
    HELLO =                            0
    SESSION_NONCE =                    1
    HEARTBEAT =                        2
    STATUS =                           3
    GET_MAC_ADDRESS =                  4
    CONTROL =                          10
    HUB_DATA_REPLY =                   11

    ENABLE_ADVERTISEMENT =             50000
    ENABLE_MESH =                      50001
    GET_CROWNSTONE_ID =                50002

    ADC_CONFIG_INC_RANGE_CURRENT =     50103
    ADC_CONFIG_DEC_RANGE_CURRENT =     50104
    ADC_CONFIG_INC_RANGE_VOLTAGE =     50105
    ADC_CONFIG_DEC_RANGE_VOLTAGE =     50106
    ADC_CONFIG_DIFFERENTIAL_CURRENT =  50108
    ADC_CONFIG_DIFFERENTIAL_VOLTAGE =  50109
    ADC_CONFIG_VOLTAGE_PIN =           50110

    POWER_LOG_CURRENT =                50200
    POWER_LOG_VOLTAGE =                50201
    POWER_LOG_FILTERED_CURRENT =       50202
    POWER_LOG_CALCULATED_POWER =       50204

    MOCK_INTERNAL_EVT =                60000
    UNKNOWN =                          65535

class UartRxType(IntEnum):
    HELLO =                            0
    SESSION_NONCE =                    1
    HEARTBEAT =                        2
    STATUS =                           3
    MAC_ADDRESS =                      4
    RESULT_PACKET =                    10

    ERR_REPLY_PARSING_FAILED =         9900
    ERR_REPLY_STATUS =                 9901
    ERR_REPLY_SESSION_NONCE_MISSING =  9902
    ERR_REPLY_DECRYPTION_FAILED =      9903

    UART_MESSAGE =                     10000
    SESSION_NONCE_MISSING =            10001
    OWN_SERVICE_DATA =                 10002
    PRESENCE_CHANGE =                  10004
    FACTORY_RESET =                    10005
    BOOTED =                           10006
    HUB_DATA =                         10007
    MICROAPP_DATA =                    10008

    MESH_SERVICE_DATA =                10102
    EXTERNAL_STATE_PART_0 =			   10103
    EXTERNAL_STATE_PART_1 =			   10104
    MESH_RESULT = 					   10105
    MESH_ACK_ALL_RESULT = 		       10106
    RSSI_PING_MESSAGE =                10107

    ASSET_MAC_RSSI_REPORT =            10108

    NEIGHBOUR_RSSI =                   10111
    ASSET_ID_RSSI_REPORT =             10112

    LOG =                              10200
    LOG_ARRAY =                        10201

    INTERNAL_EVENT =                   40000

    MESH_CMD_TIME =                    40103
    MESH_PROFILE_LOCATION =            40110
    MESH_SET_BEHAVIOUR_SETTINGS =      40111
    MESH_TRACKED_DEVICE_REGISTER =     40112
    MESH_TRACKED_DEVICE_TOKEN =        40113
    MESH_SYNC_REQUEST =                40114
    MESH_TRACKED_DEVICE_HEARTBEAT =    40120

    ADVERTISING_ENABLED =              50000
    MESH_ENABLED =                     50001
    CROWNSTONE_ID =                    50002

    ADC_CONFIG =                       50100
    ADC_RESTART =                      50101

    POWER_LOG_CURRENT =                50200
    POWER_LOG_VOLTAGE =                50201
    POWER_LOG_FILTERED_CURRENT =       50202
    POWER_LOG_FILTERED_VOLTAGE =       50203
    POWER_LOG_POWER =                  50204

    ASCII_LOG =               		   60000
    FIRMWARESTATE =                    60001



import asyncio
from crownstone_uart.core.UartEventBus import UartEventBus

class BatchCollector:
    
    def __init__(self, topic= None, timeout = 15, interval = 0.05):
        self.response = None
        self.timeout = timeout
        self.interval = interval

        self.cleanupId = None
        if topic is not None:
            self.cleanupId = UartEventBus.subscribe(topic, self.collect)

    def __del__(self):
        UartEventBus.unsubscribe(self.cleanupId)

    def cleanup(self):
        UartEventBus.unsubscribe(self.cleanupId)

    def clear(self):
        self.response = None

    async def receive(self):
        counter = 0
        while counter < self.timeout:
            if self.response is not None:
                return self.response

            await asyncio.sleep(self.interval)
            counter += self.interval
        return None


    def collect(self, data):
        self.response = data





class DevTopics:
    newServiceData = "newServiceData"  # data is dictionary: {
                                       #                        "id": int
                                       #                        "opCode": int
                                       #                        "dataType": int
                                       #                        "switchState": int
                                       #                        "flagBitmask": int
                                       #                        "temperature": int
                                       #                        "powerFactor": int
                                       #                        "powerUsageReal": int
                                       #                        "energyUsed": int
                                       #                        "partialTimestamp": int
                                       #                        "validation": int
                                       #                     }
    newCurrentData = "newCurrentData"  # data is dictionary: { id: int, type: 'current', data: [(time, data point)] }
    newVoltageData = "newVoltageData"  # data is dictionary: { id: int, type: 'voltage', data: [(time, data point)] }
    newFilteredCurrentData = "newFilteredCurrentData"  # data is dictionary: { id: int, type: 'current', data: [(time, data point)] }
    newFilteredVoltageData = "newFilteredVoltageData"  # data is dictionary: { id: int, type: 'voltage', data: [(time, data point)] }
    newCalculatedPowerData = "newCalculatedPowerData"  # data is dictionary: {
                                                       #                         "id": int,
                                                       #                         "currentRmsMA": int,
                                                       #                         "currentRmsMedianMA": int,
                                                       #                         "filteredCurrentRmsMA": int,
                                                       #                         "filteredCurrentRmsMedianMA": int,
                                                       #                         "avgZeroVoltage": int,
                                                       #                         "avgZeroCurrent": int,
                                                       #                         "powerMilliWattApparent": int,
                                                       #                         "powerMilliWattReal": int,
                                                       #                         "avgPowerMilliWattReal": int
                                                       #                     }
    newAdcConfigPacket = 'newAdcConfigPacket'
    adcRestarted = 'adcRestarted'
    ownMacAddress = 'ownMacAddress'
    ownCrownstoneId = 'ownCrownstoneId'



import logging

from crownstone_core.Exceptions import CrownstoneException
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.CRC import crc16ccitt
from crownstone_core.util.Conversion import Conversion
from crownstone_uart.core.uart.UartTypes import UartMessageType

PROTOCOL_MAJOR = 1
PROTOCOL_MINOR = 0

ESCAPE_TOKEN = 0x5c
BIT_FLIP_MASK = 0x40
START_TOKEN = 0x7e

# Size of size field.
SIZE_HEADER_SIZE = 2

# Wrapper header size: all header fields after size field.
WRAPPER_HEADER_SIZE = 1 + 1 + 1

# Wrapper crc size
CRC_SIZE = 2

_LOGGER = logging.getLogger(__name__)

class UartWrapperPacket:
	"""
	Packet:
	1B start token
	---- Size field ----
	2B size
	---- Wrapper header ----
	1B protocol major
	1B protocol minor
	1B message type
	---- Payload ----
	xB payload
	---- Tail ----
	2B CRC
	"""

	def __init__(self, messageType: UartMessageType = UartMessageType.UART_MESSAGE, payload = None):
		"""
		:param messageType:  UartMessageType
		:param payload:      List of uint8, according to messageType
		"""
		self.messageType = messageType
		self.protocolMajor = PROTOCOL_MAJOR
		self.protocolMinor = PROTOCOL_MINOR
		if payload is None:
			self.payload = []
		else:
			self.payload = payload

	def escapeCharacters(self, payload):
		escapedPayload = []
		for byte in payload:
			if byte is ESCAPE_TOKEN or byte is START_TOKEN:
				escapedPayload.append(ESCAPE_TOKEN)
				escapedByte = byte ^ BIT_FLIP_MASK
				escapedPayload.append(escapedByte)
			else:
				escapedPayload.append(byte)

		return escapedPayload

	def parse(self, buffer: list):
		"""
		Parses data between size field and CRC.

		:returns True on success.
		"""
		reader = BufferReader(buffer)
		try:
			self.protocolMajor = reader.getUInt8()
			self.protocolMinor = reader.getUInt8()
			self.messageType   = reader.getUInt8()
			self.payload       = reader.getRemainingBytes()
			return True
		except CrownstoneException as e:
			_LOGGER.warning(F"Parse error: {e}")
			return False

	def serialize(self):
		"""
		Adds start token, header and CRC to payload. Escapes bytes.
		"""
		# Construct the packet for CRC calculation
		packet = []
		packet.append(self.protocolMajor)
		packet.append(self.protocolMinor)
		packet.append(int(self.messageType))
		packet += self.payload

		# Calculate the CRC of the packet
		packetCrc = crc16ccitt(packet)

		# Append the CRC to the base packet to escape the entire thing
		packet += Conversion.uint16_to_uint8_array(packetCrc)

		# Prepend the size field.
		sizeField = len(packet)
		packet = Conversion.uint16_to_uint8_array(sizeField) + packet

		# Escape everything except the start token
		packet = self.escapeCharacters(packet)

		# Prepend the start token.
		packet = [START_TOKEN] + packet
		return packet




# voltage is channel 0
# current is channel 1
from crownstone_core.util.Conversion import Conversion

from crownstone_uart.core.uart.uartPackets.AdcChannelPacket import AdcChannelPacket


class AdcConfigPacket:
    amountOfChannels = 2
    channelSize = 6 
    packetSize = amountOfChannels * channelSize + 1 + 4 # 4 for sampling period, 1 for count

    def __init__(self, payload):
        if len(payload) < 1:
            print("ERROR: INVALID PAYLOAD LENGTH", len(payload), payload)
            return
        
        self.channels = []
        
        index = 0
        self.amountOfChannels = payload[index]
        index += 1
        self.packetSize = self.amountOfChannels * self.channelSize + 1 + 4 # 4 for sampling period, 1 for count
        
        if len(payload) < self.packetSize:
            print("ERROR: INVALID PAYLOAD LENGTH", len(payload), payload)
            return
        
        for i in range(0,self.amountOfChannels):
            self.channels.append(AdcChannelPacket(payload[ index : index+self.channelSize ], i))
            index += self.channelSize
        
        self.samplingPeriod = Conversion.uint8_array_to_uint32(payload[index:index+4])



    def getDict(self):
        data = {}
        
        data["amountOfChannels"] = self.amountOfChannels
        data["channels"] = {}
        for i in range(0, self.amountOfChannels):
            type = 'unknown'
            if i == 0:
                type = 'voltage'
            elif i == 1:
                type = 'current'
            data["channels"][type] = self.channels[i].getDict()
            
        data["samplingPeriod"] = self.samplingPeriod
        
        return data


#!/usr/bin/env python3

"""An example that sets the current time in the Crownstone mesh."""
import asyncio
import time

from crownstone_core.util.Timestamp import getCorrectedLocalTimestamp

from crownstone_uart import CrownstoneUart

uart = CrownstoneUart()

# Start up the USB bridge.
async def run_example():
	# create a connection with the crownstone usb dongle.
	await uart.initialize_usb()

	# In the Crownstone app, we usually set the local time, which is the timestamp with correction for the timezone
	# The only important thing is that you use the same timezone when you set certain time-related things as you use here.
	timestamp = time.time()

	local_timestamp = getCorrectedLocalTimestamp(timestamp)
	formattedtime = time.strftime("%Y/%m/%d %H:%M:%S",time.gmtime(local_timestamp))
	print(F"setting time to: {local_timestamp}, (i.e. {formattedtime})")
	
	await uart.mesh.set_time(int(local_timestamp))

	# stop the connection to the dongle
	uart.stop()
	

asyncio.run(run_example())



import threading
import time
import asyncio

"""
An example that switches a Crownstone, and prints the power usage of the selected Crownstone.

But now it runs it its own thread.
"""

from crownstone_uart import CrownstoneUart

# This is the id of the Crownstone we will be switching
# change it to match the Crownstone Id you want to switch!
targetCrownstoneId = 38


class UartThreadExample(threading.Thread):

    def __init__(self):
        self.loop = None
        self.uart = None
        threading.Thread.__init__(self)

    def run(self):
        self.loop = asyncio.new_event_loop()
        self.uart = CrownstoneUart()
        # choose either sync, or async operation
        self.loop.run_until_complete(self.runIt())
        # self.runIt_sync()


    async def runIt(self):
        await self.uart.initialize_usb()
        self.switch_crownstone()

    def runIt_sync(self):
        self.uart.initialize_usb_sync()
        self.switch_crownstone()

    def switch_crownstone(self):
        turnOn = True
        for i in range(0, 10):
            if not self.uart.running:
                break

            if turnOn:
                print("Switching Crownstone on  (iteration: ", i,")")
            else:
                print("Switching Crownstone off (iteration: ", i,")")
            self.uart.switch_crownstone(targetCrownstoneId, on = turnOn)
            turnOn = not turnOn
            time.sleep(2)

    def stop(self):
        self.uart.stop()


thread = UartThreadExample()
thread.start()

try:
    while 1:
        print("tick")
        time.sleep(1)
except KeyboardInterrupt:
    print("\nClosing example.... Thanks for your time!")

thread.stop()


import logging

from crownstone_core import Conversion
from crownstone_core.protocol.ControlPackets import ControlPacketsGenerator

from crownstone_uart.core.modules.ControlHandler import ControlHandler

_LOGGER = logging.getLogger(__name__)

class SetupHandler:
    def __init__(self, control: ControlHandler):
        self.control = control
        pass

    async def setup(self,
                    sphereId: int,
                    crownstoneId: int,
                    meshDeviceKey: bytearray or list,
                    ibeaconUUID: str,
                    ibeaconMajor: int,
                    ibeaconMinor: int,
                    keysDict: dict):
        """
        Perform the setup of the Crownstone we are connected to via UART.
        
        :param sphereId:      An ID (0-255) that is the same for each Crownstone within the sphere.
        :param crownstoneId:  An ID (1-255) that is unique within the sphere.
        :param meshDeviceKey: A 16 byte key that is different for each Crownstone.
        :param ibeaconUUID:   The iBeacon UUID. Should be the same for each Crownstone in the sphere.
                              Provided as string, like: "1843423e-e175-4af0-a2e4-31e32f729a8a".
        :param ibeaconMajor:  The iBeacon major. Together with the minor, should be unique per sphere.
        :param ibeaconMinor:  The iBeacon minor. Together with the major, should be unique per sphere.
        :param keysDict:      Dictionary with the sphere keys.
        """
        controlPacket = ControlPacketsGenerator.getSetupPacket(
            crownstoneId,
            sphereId,
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["admin"]),
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["member"]),
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["basic"]),
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["serviceDataKey"]),
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["localizationKey"]),
            meshDeviceKey,
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["meshApplicationKey"]),
            Conversion.ascii_or_hex_string_to_16_byte_array(keysDict["meshNetworkKey"]),
            ibeaconUUID,
            ibeaconMajor,
            ibeaconMinor
        )
        await self.control._writeControlAndWaitForSuccess(controlPacket)



import asyncio, logging, time
from typing import List

from crownstone_core.Exceptions import CrownstoneException
from crownstone_core.packets.ResultPacket import ResultPacket
from crownstone_core.protocol.BluenetTypes import ResultValue

from crownstone_uart.Constants import UART_WRITE_TIMEOUT
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.topics.SystemTopics import SystemTopics

_LOGGER = logging.getLogger(__name__)

"""
This class will handle the event flow around writing to uart and receiving errors or result codes.

Usage of code duplication is because asyncio is difficult when obtaining event loops. You cant get one if you're already in async method, etc.
To avoid this annoying behaviour, we duplicate the code a bit.
"""
class UartWriter:
    def __init__(self, dataToSend: List[int], interval = 0.001):
        """
        This class will handle the event flow around writing to uart and receiving errors or result codes.
        :param dataToSend: This is your data packet
        :param interval: Polling interval. Don't touch. This is cheap and local only. It does not do uart things.
        """
        self.dataToSend : List[int] = dataToSend
        self.interval = interval

        self.error   = None
        self.success = False
        self.result  = None

        self.cleanupIds = []
        self.cleanupIds.append(UartEventBus.subscribe(SystemTopics.resultPacket,     self._handleResult))
        self.cleanupIds.append(UartEventBus.subscribe(SystemTopics.uartWriteSuccess, self._handleSuccess))
        self.cleanupIds.append(UartEventBus.subscribe(SystemTopics.uartWriteError,   self._handleError))

    def __del__(self):
        for cleanupId in self.cleanupIds:
            UartEventBus.unsubscribe(cleanupId)

    def _handleError(self, errorData):
        _LOGGER.error("Error during uart write", errorData)
        self.error = errorData
        raise errorData["error"]

    def _handleResult(self, data: ResultPacket):
        self.result = data
        _LOGGER.debug("Uart result packet received")

    def _handleSuccess(self, data: List[int]):
        if data == self.dataToSend:
            _LOGGER.debug("Uart write successful")
            self.success = True

    async def write_with_result(self, success_codes=None, result_timeout=1) -> ResultPacket:
        """
        write_with_result will take the data packet you have provided to the constructor and send it over UART.
        It will wait for an incoming result packet and return it once it has been received.

        You can optionally provide a list of ResultValue result codes. If the list is empty [] all resultPackets will be returned.
        Otherwise it will use your provided list to filter on. If you leave it None, we will default to ResultValue.SUCCESS.

        If a resultcode is not in the list, an CrownstoneException will be raised. You can await the result. If the write fails, an CrownstoneException will be raised.

        :param success_codes: List[ResultValue]
        :param result_timeout: time to wait on a result in seconds before raising a timeout CrownstoneException. This should be higher than the UART_WRITE_TIMEOUT
        :return: ResultPacket
        """

        if success_codes is None:
            success_codes = [ResultValue.SUCCESS]
        UartEventBus.emit(SystemTopics.uartWriteData, self.dataToSend)
        counter = 0
        while counter < result_timeout:
            if self.result:
                return self._checkResult(success_codes)

            await asyncio.sleep(self.interval)
            counter += self.interval

        self._wrapUpFailedResult(result_timeout)

    def write_with_result_sync(self, success_codes : List[ResultValue]=None, result_timeout=0.25) -> ResultPacket:
        """
        write_with_result_sync will take the data packet you have provided to the constructor and send it over UART.
        It will wait for an incoming result packet and return it once it has been received.

        You can optionally provide a list of ResultValue result codes. If the list is empty [] all resultPackets will be returned.
        Otherwise it will use your provided list to filter on. If you leave it None, we will default to ResultValue.SUCCESS.

        If a resultcode is not in the list, an CrownstoneException will be raised. This method is blocking.
        If the write fails, an CrownstoneException will be raised.
        :return:

        :param success_codes: List[ResultValue]
        :param result_timeout: time to wait on a result in seconds before raising a timeout error. This should be higher than the UART_WRITE_TIMEOUT
        :return: ResultPacket
        """
        if success_codes is None:
            success_codes = [ResultValue.SUCCESS]
        UartEventBus.emit(SystemTopics.uartWriteData, self.dataToSend)
        counter = 0
        while counter < result_timeout:
            if self.result:
                return self._checkResult(success_codes)

            time.sleep(self.interval)
            counter += self.interval

        self._wrapUpFailedResult(result_timeout)

    async def write(self) -> True:
        """
        write will take the data packet you have provided to the constructor and send it over UART.
        You can await the success of the write. If the write fails, an CrownstoneException will be raised.
        :return:
        """

        UartEventBus.emit(SystemTopics.uartWriteData, self.dataToSend)
        counter = 0
        while counter < 2*UART_WRITE_TIMEOUT:
            if self.success:
                # cleanup the listener(s)
                self.__del__()
                return True

            await asyncio.sleep(self.interval)
            counter += self.interval


        self._wrapUpFailedWrite()
    
    def write_sync(self) -> True:
        """
        write_sync will take the data packet you have provided to the constructor and send it over UART.
        This method is blocking. If the write fails, an CrownstoneException will be raised.
        :return:
        """

        UartEventBus.emit(SystemTopics.uartWriteData, self.dataToSend)
        counter = 0
        while counter < 2 * UART_WRITE_TIMEOUT:
            if self.success:
                # cleanup the listener(s)
                self.__del__()
                return True

            time.sleep(self.interval)
            counter += self.interval
        self._wrapUpFailedWrite()



    def _checkResult(self, success_codes):
        if success_codes == [] or self.result.resultCode in success_codes:
            self.__del__()
            return self.result
        else:
            # TODO: Invalid error type.
            raise CrownstoneException(
                "WRITE_EXCEPTION",
                "Incorrect result type. Got " + str(self.result.resultCode) + ", expected one of " + str(success_codes),
                400
            )


    def _wrapUpFailedResult(self, wait_until_result):
        self.__del__()
        # TODO: Invalid error type.
        raise CrownstoneException("WRITE_EXCEPTION", "No result received after writing to UART. Waited for " + str(
            wait_until_result) + " seconds", 404)


    def _wrapUpFailedWrite(self):
        # we should never arrive here. If the write went wrong, an error should have been thrown by the write method before we get here.
        self.__del__()
        # TODO: Invalid error type.
        raise CrownstoneException("WRITE_EXCEPTION", "Write not completed, but no error was thrown.", 500)



#!/usr/bin/env python3

"""
Example to upload asset filters via UART.
"""
import asyncio
import logging

import crownstone_core
from crownstone_core.packets.assetFilter.FilterMetaDataPackets import FilterType
from crownstone_core.packets.assetFilter.builders.AssetFilter import AssetFilter

from crownstone_uart import CrownstoneUart, UartEventBus, UartTopics
from crownstone_uart.core.uart.uartPackets.AssetIdReport import AssetIdReport
from crownstone_uart.core.uart.uartPackets.AssetMacReport import AssetMacReport

##### Settings #####

# The uart port to use.
# Can be set to /dev/ttyUSB0 for example.
port = None

# Set to true for debug logs.
verbose = False


if verbose:
	logging.basicConfig(format='%(asctime)s %(levelname)-7s: %(message)s', level=logging.DEBUG)

# Init the Crownstone UART lib.
uart = CrownstoneUart()

print("core version:", crownstone_core.__version__)
print("uart version:", uart.__version__)

async def main():
	# The try except part is just to catch a control+c to gracefully stop the libs.
	try:
		await uart.initialize_usb(port=port, writeChunkMaxSize=64)

		def onAssetMac(report: AssetMacReport):
			print(f"onAssetMac mac={report.assetMacAddress} crownstoneId={report.crownstoneId} rssi={report.rssi} channel={report.channel}")

		def onAssetId(report: AssetIdReport):
			print(f"onAssetId id={report.assetId} crownstoneId={report.crownstoneId} rssi={report.rssi} channel={report.channel} passedFilterIds={report.passedFilterIds}")

		UartEventBus.subscribe(UartTopics.assetTrackingReport, onAssetMac)
		UartEventBus.subscribe(UartTopics.assetIdReport, onAssetId)

		filterId = 0
		filter1 = AssetFilter(filterId)
		filter1.filterByMacAddress(["01:23:45:67:89:AB", "01:23:45:67:89:CD"])
		filter1.outputMacRssiReport()
		filter1.setProfileId(0)
		print("filter1:")
		print(filter1)

		filterId += 1
		filter2 = AssetFilter(filterId)
		filter2.filterByNameWithWildcards("C?W*", complete=False)
		filter2.outputAssetId().basedOnMac()
		filter2.setFilterType(FilterType.CUCKOO)
		print("filter2:")
		print(filter2)

		filters = [filter1, filter2]

		print("Set filters via UART.")
		masterVersion = await uart.control.setFilters(filters)
		print("Master version:", masterVersion)

		# Simply keep the program running.
		while True:
			await asyncio.sleep(0.1)
	except KeyboardInterrupt:
		pass
	finally:
		print("\nStopping UART..")
		uart.stop()
		print("Stopped")

asyncio.run(main())



from crownstone_core.util.Conversion import Conversion


class PowerCalculationPacket:
	amountOfSamples = 9
	sampleSize = 4
	packetSize = amountOfSamples * sampleSize

	def __init__(self, payload):
		self.currentRmsMA = 0
		self.currentRmsMedianMA = 0
		self.filteredCurrentRmsMA = 0
		self.filteredCurrentRmsMedianMA = 0
		self.avgZeroVoltage = 0
		self.avgZeroCurrent = 0
		self.powerMilliWattApparent = 0
		self.powerMilliWattReal = 0
		self.avgPowerMilliWattReal = 0
		
		if len(payload) < self.packetSize:
			print("ERROR: INVALID PAYLOAD LENGTH", len(payload), payload)
			return

		i = 0
		self.currentRmsMA 				= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.currentRmsMedianMA 		= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.filteredCurrentRmsMA 		= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.filteredCurrentRmsMedianMA = Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.avgZeroVoltage 			= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.avgZeroCurrent 			= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.powerMilliWattApparent 	= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.powerMilliWattReal 		= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])
		i += self.sampleSize

		self.avgPowerMilliWattReal 		= Conversion.uint8_array_to_int32(payload[i:i+self.sampleSize])

	def getDict(self):
		data = {}

		data["id"] = 0  # TODO: get the Crownstone ID here.
		data["currentRmsMA"] = self.currentRmsMA
		data["currentRmsMedianMA"] = self.currentRmsMedianMA
		data["filteredCurrentRmsMA"] = self.filteredCurrentRmsMA
		data["filteredCurrentRmsMedianMA"] = self.filteredCurrentRmsMedianMA
		data["avgZeroVoltage"] = self.avgZeroVoltage
		data["avgZeroCurrent"] = self.avgZeroCurrent
		data["powerMilliWattApparent"] = self.powerMilliWattApparent
		data["powerMilliWattReal"] = self.powerMilliWattReal
		data["avgPowerMilliWattReal"] = self.avgPowerMilliWattReal

		return data


import asyncio
import time
import logging
from queue import Queue
from typing import Callable, Any

from crownstone_uart.core.UartEventBus import UartEventBus

_LOGGER = logging.getLogger(__name__)

class Collector:
    """
    Class that waits for data from the given topic.
    """
    def __init__(self, topic= None, timeout = 10, interval = 0.05):
        """
        :param topic:        The topic to receive data from.
        :param timeout:      Timeout in seconds.
        :param interval:     Polling interval.
        """
        # Config variables
        self.topic = topic
        self.timeout = timeout
        self.interval = interval

        # State variables
        self.response = None
        self.responses = Queue()
        self.timeLeft = -1
        self.callback = None
        self.cleanupId = None

    def __del__(self):
        self._cleanup()

    def _start(self):
        self.response = None
        self.responses = Queue()
        self.timeLeft = self.timeout
        if self.topic is not None:
            self.cleanupId = UartEventBus.subscribe(self.topic, self._onData)
        _LOGGER.debug(f"Start result collector with timeLeft={self.timeLeft} and topic={self.topic}")

    def _cleanup(self):
        _LOGGER.debug(f"Cleanup result collector")
        if self.cleanupId is not None:
            UartEventBus.unsubscribe(self.cleanupId)

        # Don't clean up response yet, they are used as return data.
        self.timeLeft = -1
        self.callback = None
        self.cleanupId = None

    # def clear(self):
    #     self.response = None

    async def receive(self):
        """
        Wait for data from the given topic.

        :return: The received data, in the format specified by the topic, or None on timeout.
        """
        self._start()
        while self.timeLeft > 0:
            if self.response is not None:
                self._cleanup()
                return self.response

            await asyncio.sleep(self.interval)
            self.timeLeft -= self.interval

        self._cleanup()
        return None

    def receive_sync(self):
        """
        Wait for data from the given topic.

        :return: The received data, in the format specified by the topic.
        """
        self._start()
        while self.timeLeft > 0:
            if self.response is not None:
                self._cleanup()
                return self.response

            time.sleep(self.interval)
            self.timeLeft -= self.interval

        self._cleanup()
        return None

    async def receiveNext(self):
        """
        Wait for data from the given topic.

        :return: The received data, in the format specified by the topic, or None on timeout.
        """
        if self.cleanupId is None:
            self._start()

        while self.timeLeft > 0:
            if not self.responses.empty():
                return self.responses.get()

            await asyncio.sleep(self.interval)
            self.timeLeft -= self.interval

        self._cleanup()
        return None

    async def receiveMultiple(self, callback: Callable[[Any], bool]):
        """
        Wait for data and let the callback handle the data.
        :param callback:          Function that handles the data and returns True to continue collecting, or False to stop collecting.
        """
        self._start()
        self.callback = callback
        await asyncio.sleep(self.timeLeft)
        callback(None)
        self._cleanup()

    def _onData(self, data):
        _LOGGER.debug(f"Result collector received {data}")
        if self.callback is None:
            self.response = data
            self.responses.put(data)
        else:
            if not self.callback(data):
                self._cleanup()

        # To be sure: always clean up when timed out.
        if self.timeLeft < 0:
            self._cleanup()





# voltage is channel 0
# current is channel 1
from crownstone_core.util.Conversion import Conversion


class AdcChannelPacket:
    packetSize = 6

    def __init__(self, payload, channelIndex):
        if len(payload) < self.packetSize:
            print("ERROR: INVALID PAYLOAD LENGTH", len(payload), payload)
            return
        
        self.channelIndex = channelIndex
        
        self.pin    = payload[0]
        self.range  = Conversion.uint8_array_to_uint32(payload[1:1+4])
        self.refPin = payload[5]

    def getDict(self):
        data = {}

        data["pin"] = self.pin
        data["range"] = self.range
        data["refPin"] = self.refPin
        data["channelIndex"] = self.channelIndex

        return data


import asyncio

from crownstone_core.Exceptions import CrownstoneException
from crownstone_core.protocol.BlePackets import ControlStateSetPacket
from crownstone_core.protocol.BluenetTypes import StateType, ResultValue

from crownstone_uart.core.dataFlowManagers.Collector import Collector
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket
from crownstone_uart.core.uart.UartTypes import UartTxType, UartMessageType
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket
from crownstone_uart.topics.SystemTopics import SystemTopics


class StateHandler:

    async def setPowerZero(self, mW: int):
        controlPacket = ControlStateSetPacket(StateType.POWER_ZERO).loadInt32(mW).serialize()
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        resultCollector = Collector(timeout=1,  topic=SystemTopics.resultPacket)
        # send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        # wait for the collectors to fill
        commandResultData = await resultCollector.receive()

        if commandResultData is not None:
            if commandResultData.resultCode is ResultValue.BUSY:
                await asyncio.sleep(0.2)
                return await self.setPowerZero(mW)
            elif commandResultData.resultCode is not ResultValue.SUCCESS:
                raise CrownstoneException(commandResultData.resultCode, "Command has failed.")



from crownstone_core.protocol.BlePackets import ControlStateSetPacket, ControlPacket
from crownstone_core.protocol.BluenetTypes import StateType, ControlType

from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket
from crownstone_uart.core.uart.UartTypes import UartTxType, UartMessageType
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket
from crownstone_uart.core.dataFlowManagers.UartWriter import UartWriter
from crownstone_uart.topics.SystemTopics import SystemTopics

from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.packets.microapp.MicroappInfoPacket import MicroappInfoPacket
from crownstone_core.packets.microapp.MicroappUploadPacket import MicroappUploadPacket

class UsbDevHandler:
    
    def setAdvertising(self, enabled):
        """
            Enable/ disable the advertising
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.ENABLE_ADVERTISEMENT, self._getPayload(enabled))
    
    def setMeshing(self, enabled):
        """
            Enable/ disable the Meshing
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.ENABLE_MESH, self._getPayload(enabled))
    
    def requestCrownstoneId(self):
        """
            Request the Crownstone ID. This is a uint16
            :return:
        """
        self._send(UartTxType.GET_CROWNSTONE_ID, [])
    
    def requestMacAddress(self):
        """
            Request the MAC address ID.
            :return:
        """
        self._send(UartTxType.GET_MAC_ADDRESS, [])
    
    def increaseCurrentRange(self):
        """
            Increase the GAIN on the current sensing
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_INC_RANGE_CURRENT, [])
    
    def decreaseCurrentRange(self):
        """
            Decrease the GAIN on the current sensing
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_DEC_RANGE_CURRENT, [])
    
    def increaseVoltageRange(self):
        """
            Increase the GAIN on the voltage sensing
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_INC_RANGE_VOLTAGE, [])
    
    def decreaseVoltageRange(self):
        """
            Decrease the GAIN on the voltage sensing
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_DEC_RANGE_VOLTAGE, [])
    
    def setDifferentialModeCurrent(self, enabled):
        """
            Enable/disable differential mode on the current sensing
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_DIFFERENTIAL_CURRENT, self._getPayload(enabled))
    
    def setDifferentialModeVoltage(self, enabled):
        """
            Enable/disable differential mode on the voltage sensing
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_DIFFERENTIAL_VOLTAGE, self._getPayload(enabled))

    def setVoltageChannelPin(self, pin):
        """
            Select the measurement pin for the voltage sensing
            :param pin: int [0 .. 255]
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_VOLTAGE_PIN, [pin])

    def toggleVoltageChannelPin(self):
        """
            Select the measurement pin for the voltage sensing
            :return:
        """
        self._send(UartTxType.ADC_CONFIG_VOLTAGE_PIN, [])

    def setSendCurrentSamples(self, enabled):
        """
            Enable/ disable the sending of the measured current buffer.
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.POWER_LOG_CURRENT, self._getPayload(enabled))

    def setSendVoltageSamples(self, enabled):
        """
            Enable/ disable the sending of the measured voltage buffer.
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.POWER_LOG_VOLTAGE, self._getPayload(enabled))
    
    def setSendFilteredCurrentSamples(self, enabled):
        """
            Enable/ disable the sending of the filtered current sample buffer.
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.POWER_LOG_FILTERED_CURRENT, self._getPayload(enabled))
     
    def setSendFilteredVoltageSamples(self, enabled):
        """
            Enable/ disable the sending of the filtered voltage sample buffer.
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.POWER_LOG_FILTERED_VOLTAGE, self._getPayload(enabled))
    
    def setSendCalculatedSamples(self, enabled):
        """
            Enable/ disable the sending of the calculated power samples.
            :param enabled: Boolean
            :return:
        """
        self._send(UartTxType.POWER_LOG_CALCULATED_POWER, self._getPayload(enabled))

    def setUartMode(self, mode):
        """
            Set UART mode.
            :param mode: : 0=none 1=RX only, 2=TX only, 3=TX and RX
            :return:
        """
        if (mode < 0) or (mode > 3):
            return
        controlPacket = ControlStateSetPacket(StateType.UART_ENABLED).loadUInt8(mode).serialize()
        self._send(UartTxType.CONTROL, controlPacket)

    def resetCrownstone(self):
        """
            Reset the Crownstone
            :return:
        """
        resetPacket = ControlPacket(ControlType.RESET).serialize()
        self._send(UartTxType.CONTROL, resetPacket)

    def toggleRelay(self, isOn):
        val = 0
        if isOn:
            val = 1
        switchPacket = ControlPacket(ControlType.RELAY).loadUInt8(val).serialize()
        self._send(UartTxType.CONTROL, switchPacket)

    def toggleIGBTs(self, isOn):
        val = 0
        if isOn:
            val = 100
        switchPacket = ControlPacket(ControlType.PWM).loadUInt8(val).serialize()
        self._send(UartTxType.CONTROL, switchPacket)

    def toggleAllowDimming(self, isOn):
        val = 0
        if isOn:
            val = 1
        instructionPacket = ControlPacket(ControlType.ALLOW_DIMMING).loadUInt8(val).serialize()
        self._send(UartTxType.CONTROL, instructionPacket)

    def remove_microapp(self, index : int):
        """
        Remove microapp from flash memory in Bluenet. Return True if message is send correctly, it
        doesn't check if the command is set correctly in Bluenet. For now only an index of 0 can
        be given, since Bluenet only supports a single app to run.
        """
        packet = MicroappHeaderPacket(index)
        controlPacket = ControlPacket(
            ControlType.MICROAPP_REMOVE).loadByteArray(packet.serialize()).serialize()
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartWriter(uartPacket).write_sync()

    def enable_microapp(self, index : int):
        """
        Enable microapp in Bluenet and load it in RAM. Return True if message is send correctly,
        it doesn't check if the command is set correctly in Bluenet. For now only an index of 0
        can be given, since Bluenet only supports a single app to run. It is recommended to run
        the validation function before enabling the microapp to make sure a microapp is able to
        run.
        """
        packet = MicroappHeaderPacket(index)
        controlPacket = ControlPacket(
            ControlType.MICROAPP_ENABLE).loadByteArray(packet.serialize()).serialize()
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartWriter(uartPacket).write_sync()

    def validate_microapp(self, index : int):
        """
        validate the binary of a microapp saved in Bluenet flash memory. Return True if message is
        send correctly, it doesn't check if the command is set correctly in Bluenet. For now only
        an index of 0 can be given, since Bluenet only supports a single app to run.
        """
        packet = MicroappHeaderPacket(index)
        controlPacket = ControlPacket(
            ControlType.MICROAPP_VALIDATE).loadByteArray(packet.serialize()).serialize()
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartWriter(uartPacket).write_sync()

    def disable_microapp(self, index : int):
        """
        Disable a running microapp in Bluenet by removing it from RAM. Return True if message is
        send correctly, it doesn't check if the command is set correctly in Bluenet. For now only
        an index of 0 can be given, since Bluenet only supports a single app to run.
        """
        packet = MicroappHeaderPacket(index)
        controlPacket = ControlPacket(
            ControlType.MICROAPP_DISABLE).loadByteArray(packet.serialize()).serialize()
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartWriter(uartPacket).write_sync()



    def _getPayload(self, boolean):
        payload = 0
        if boolean:
            payload = 1
            
        return [payload]

    def _send(self, opCode: UartTxType, payload: list):
        # send over uart
        uartMessage = UartMessagePacket(opCode, payload).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)



from crownstone_uart.core.uart.uartPackets.CurrentSamplesPacket import CurrentSamplesPacket


class VoltageSamplesPacket(CurrentSamplesPacket):
	def __init__(self, payload):
		super().__init__(payload)
		self.type = 'voltage'



import logging

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader

from crownstone_uart.core.uart.uartPackets.UartLogHeaderPacket import UartLogHeaderPacket

_LOGGER = logging.getLogger(__name__)

class UartLogPacket(BasePacket):
	"""
	UART log packet
	"""

	def __init__(self, data = None):
		self.header = UartLogHeaderPacket()
		self.numArgs = 0

		# List of buffers, one for each argument.
		self.argBufs = []

		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		self.header.deserialize(reader)
		self.numArgs = reader.getUInt8()
		self.argBufs = []
		for i in range(0, self.numArgs):
			argSize = reader.getUInt8()
			self.argBufs.append(reader.getBytes(argSize))

	def __str__(self):
		return f"UartLogPacket(" \
		       f"header={self.header}, " \
		       f"numArgs={self.numArgs}, " \
		       f"argBufs={self.argBufs})"



from typing import List

from crownstone_core.protocol.BluenetTypes import ResultValue


class MeshResult:

    def __init__(self, crownstone_uid_array : List[int]):
        self.success = False
        self.resultCode = ResultValue.UNSPECIFIED
        self.acks = {}
        self.initial_crownstone_uid_array = crownstone_uid_array
        for uid in crownstone_uid_array:
            self.acks[uid] = False

    def merge(self, result: 'MeshResult'):
        for uid, success in result.acks.items():
            self.acks[uid] = result.acks[uid]

    def get_successful_ids(self) -> List[int]:
        successList = []
        for uid, success in self.acks.items():
            if self.acks[uid]:
                successList.append(uid)
        return successList

    def compare_get_failed(self, otherResult: 'MeshResult') -> List[int]:
        failedList = []
        for uid, success in self.acks.items():
            if success and uid in otherResult.acks:
                if not otherResult.acks[uid]:
                    failedList.append(uid)
        return failedList

    def compare_get_success(self, otherResult: 'MeshResult') -> List[int]:
        successList = []
        for uid, success in self.acks.items():
            if success and uid in otherResult.acks:
                if otherResult.acks[uid]:
                    successList.append(uid)
        return successList

    def collect_ack(self, uid, result):
        if uid in self.acks:
            self.acks[uid] = result

    def wasSuccessful(self, crownstone_uid_array : List[int] = None):
        array_to_use = self.initial_crownstone_uid_array
        if crownstone_uid_array is not None:
            array_to_use = crownstone_uid_array
        for uid in array_to_use:
            if not self.acks[uid]:
                return False
        return True

    def conclude(self, crownstone_uid_array : List[int] = None):
        self.success = self.wasSuccessful(crownstone_uid_array)


import datetime
import logging

from crownstone_core.Exceptions import CrownstoneException
from crownstone_core.packets.ResultPacket import ResultPacket
from crownstone_core.packets.ServiceData import ServiceData
from crownstone_core.packets.microapp.MicroappMessagePacket import MicroappMessagePacket
from crownstone_core.packets.serviceDataParsers.parsers import parseOpcode7
from crownstone_core.util.Conversion import Conversion

from crownstone_uart.core.uart.uartPackets.AssetIdReport import AssetIdReport
from crownstone_uart.core.uart.uartPackets.AssetMacReport import AssetMacReport
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.UartTypes import UartRxType, UartMessageType
from crownstone_uart.core.uart.uartPackets.AdcConfigPacket import AdcConfigPacket
from crownstone_uart.core.uart.uartPackets.CurrentSamplesPacket import CurrentSamplesPacket
from crownstone_uart.core.uart.uartPackets.PowerCalculationPacket import PowerCalculationPacket
from crownstone_uart.core.uart.uartPackets.RssiBetweenStonesPacket import RssiBetweenStonesPacket
from crownstone_uart.core.uart.uartPackets.UartCrownstoneHelloPacket import UartCrownstoneHelloPacket
from crownstone_uart.core.uart.uartPackets.UartLogArrayPacket import UartLogArrayPacket
from crownstone_uart.core.uart.uartPackets.UartLogPacket import UartLogPacket
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket, PROTOCOL_MAJOR
from crownstone_uart.core.uart.uartPackets.VoltageSamplesPacket import VoltageSamplesPacket
from crownstone_uart.topics.DevTopics import DevTopics
from crownstone_uart.topics.SystemTopics import SystemTopics
from crownstone_uart.topics.UartTopics import UartTopics

_LOGGER = logging.getLogger(__name__)

class UartParser:
    """
    Receives SystemTopics.uartNewPackage messages and emits their corresponding SystemTopics.uartNewMessage events.
    Several opcodes for uartNewMessage wil subsequently be parsed (looped back) and a more specific event may
    be emitted.
    """
    
    def __init__(self):
        self.uartPackageSubscription = UartEventBus.subscribe(SystemTopics.uartNewPackage, self.parse)
        self.uartMessageSubscription = UartEventBus.subscribe(SystemTopics.uartNewMessage, self.handleUartMessage)
        self.timestampFormat = "%Y-%m-%d %H:%M:%S.%f"

    def stop(self):
        UartEventBus.unsubscribe(self.uartPackageSubscription)
        UartEventBus.unsubscribe(self.uartMessageSubscription)

    def parse(self, wrapperPacket: UartWrapperPacket):
        """
        Callback for SystemTopics.uartNewPackage, emits a message of type SystemTopic.uartNewMessage
        if the received message was of the correct protocol version and unencrypted.
        :param wrapperPacket:
        :return:
        """
        if type(wrapperPacket) is not UartWrapperPacket:
            raise TypeError(f"Invalid type: {type(wrapperPacket)}")

        if wrapperPacket.protocolMajor != PROTOCOL_MAJOR:
            _LOGGER.warning(F"Unknown protocol: {wrapperPacket.protocolMajor}.{wrapperPacket.protocolMinor}")
            return

        msgType = wrapperPacket.messageType
        if msgType == UartMessageType.UART_MESSAGE:
            uartMsg = UartMessagePacket()
            if uartMsg.parse(wrapperPacket.payload):
                UartEventBus.emit(SystemTopics.uartNewMessage, uartMsg)
        elif msgType == UartMessageType.ENCRYPTED_UART_MESSAGE:
            _LOGGER.info(f"Received encrypted msg: decryption is not implemented.")
            return
        else:
            _LOGGER.warning(F"Unknown message type: {msgType}")
            return

    def handleUartMessage(self, messagePacket: UartMessagePacket):
        try:
            self._handleUartMessage(messagePacket)
        except CrownstoneException as e:
            # TODO: don't we catch too many errors this way
            #  For example errors from code executed by emitted events.
            #  We only want to catch parse errors, but we're too lazy to put a try catch around every parse call.
            _LOGGER.error(f"Parse error: {e}")

    def _handleUartMessage(self, messagePacket: UartMessagePacket):
        """
        Callback for SystemTopics.uartNewMessage. This transforms a select number of message types
        into further specialized messages and posts those on UartEventBus.
        :param messagePacket:
        :return:
        """
        opCode = messagePacket.opCode
        parsedData = None
        # print("UART - opCode:", opCode, "payload:", messagePacket.payload)

        if opCode == UartRxType.HELLO:
            helloPacket = UartCrownstoneHelloPacket(messagePacket.payload)
            UartEventBus.emit(UartTopics.hello, helloPacket)
            pass

        elif opCode == UartRxType.SESSION_NONCE:
            _LOGGER.debug(f"Received SESSION_NONCE")
            pass

        elif opCode == UartRxType.HEARTBEAT:
            _LOGGER.debug(f"Received HEARTBEAT")
            pass

        elif opCode == UartRxType.STATUS:
            _LOGGER.debug(f"Received STATUS")
            pass

        elif opCode == UartRxType.RESULT_PACKET:
            packet = ResultPacket(messagePacket.payload)
            UartEventBus.emit(SystemTopics.resultPacket, packet)


        #################
        # Error replies #
        #################

        elif opCode == UartRxType.ERR_REPLY_PARSING_FAILED:
            _LOGGER.debug(f"Received ERR_REPLY_PARSING_FAILED")
            pass

        elif opCode == UartRxType.ERR_REPLY_STATUS:
            _LOGGER.debug(f"Received ERR_REPLY_STATUS")
            pass

        elif opCode == UartRxType.ERR_REPLY_SESSION_NONCE_MISSING:
            _LOGGER.debug(f"Received ERR_REPLY_SESSION_NONCE_MISSING")
            pass

        elif opCode == UartRxType.ERR_REPLY_DECRYPTION_FAILED:
            _LOGGER.debug(f"Received ERR_REPLY_DECRYPTION_FAILED")
            pass

        elif 9900 < opCode < 10000:
            _LOGGER.debug(f"Received ERR_REPLY {opCode}")
            pass


        ################
        #### Events ####
        ################

        elif opCode == UartRxType.UART_MESSAGE:
            stringResult = ""
            for byte in messagePacket.payload:
                stringResult += chr(byte)
            # logStr = "LOG: %15.3f - %s" % (time.time(), stringResult)
            UartEventBus.emit(UartTopics.uartMessage, {"string":stringResult, "data": messagePacket.payload})

        elif opCode == UartRxType.SESSION_NONCE_MISSING:
            _LOGGER.debug(f"Received SESSION_NONCE_MISSING")
            pass

        elif opCode == UartRxType.OWN_SERVICE_DATA:
            # service data type + device type + data type + service data (15b)
            serviceData = ServiceData(messagePacket.payload)
            serviceData.parse()
            UartEventBus.emit(DevTopics.newServiceData, serviceData.payload)

        elif opCode == UartRxType.PRESENCE_CHANGE:
            pass

        elif opCode == UartRxType.FACTORY_RESET:
            pass

        elif opCode == UartRxType.BOOTED:
            _LOGGER.debug(f"Received BOOTED")
            pass

        elif opCode == UartRxType.HUB_DATA:
            pass

        elif opCode == UartRxType.MICROAPP_DATA:
            packet = MicroappMessagePacket.fromData(messagePacket.payload)
            _LOGGER.debug(f"Received microapp message: {packet}")
            UartEventBus.emit(UartTopics.microappMessage, packet)

        elif opCode == UartRxType.MESH_SERVICE_DATA:
            # data type + service data (15b)
            result = parseOpcode7(messagePacket.payload)
            _LOGGER.debug(f"Received service data: {result}")
            if hasattr(result,"crownstoneId"):
                UartEventBus.emit(SystemTopics.stateUpdate, (result.crownstoneId, result))

        elif opCode == UartRxType.EXTERNAL_STATE_PART_0:
            pass

        elif opCode == UartRxType.EXTERNAL_STATE_PART_1:
            pass

        elif opCode == UartRxType.MESH_RESULT:
            if len(messagePacket.payload) > 1:
                crownstoneId = messagePacket.payload[0]
                packet = ResultPacket(messagePacket.payload[1:])
                UartEventBus.emit(SystemTopics.meshResultPacket, [crownstoneId, packet])

        elif opCode == UartRxType.MESH_ACK_ALL_RESULT:
            packet = ResultPacket(messagePacket.payload)
            UartEventBus.emit(SystemTopics.meshResultFinalPacket, packet)

        elif opCode == UartRxType.RSSI_PING_MESSAGE:
            # for now, you can subscribe to SystemTopics.uartNewMessage
            pass

        elif opCode == UartRxType.LOG:
            _LOGGER.debug(f"Received binary log: {messagePacket.payload}")
            packet = UartLogPacket(messagePacket.payload)
            UartEventBus.emit(UartTopics.log, packet)

        elif opCode == UartRxType.LOG_ARRAY:
            _LOGGER.debug(f"Received binary log array: {messagePacket.payload}")
            packet = UartLogArrayPacket(messagePacket.payload)
            UartEventBus.emit(UartTopics.logArray, packet)

        elif opCode == UartRxType.NEIGHBOUR_RSSI:
            _LOGGER.debug(f"Received neighbour RSSI: {messagePacket.payload}")
            packet = RssiBetweenStonesPacket(messagePacket.payload)
            UartEventBus.emit(UartTopics.rssiBetweenStones, packet)

        #######################
        # ASSET FILTER events #
        #######################

        elif opCode == UartRxType.ASSET_MAC_RSSI_REPORT:
            _LOGGER.debug(f"Received ASSET_MAC_RSSI_REPORT: {messagePacket.payload}")
            packet = AssetMacReport(messagePacket.payload)
            UartEventBus.emit(UartTopics.assetTrackingReport, packet)

        elif opCode == UartRxType.ASSET_ID_RSSI_REPORT:
            _LOGGER.debug(f"Received ASSET_ID_RSSI_REPORT: {messagePacket.payload}")
            packet = AssetIdReport(messagePacket.payload)
            UartEventBus.emit(UartTopics.assetIdReport, packet)


        ####################
        # Developer events #
        ####################

        elif opCode == UartRxType.INTERNAL_EVENT:
            pass

        elif opCode == UartRxType.MESH_CMD_TIME:
            pass

        elif opCode == UartRxType.MESH_PROFILE_LOCATION:
            pass

        elif opCode == UartRxType.MESH_SET_BEHAVIOUR_SETTINGS:
            pass

        elif opCode == UartRxType.MESH_TRACKED_DEVICE_REGISTER:
            pass

        elif opCode == UartRxType.MESH_TRACKED_DEVICE_TOKEN:
            pass

        elif opCode == UartRxType.MESH_SYNC_REQUEST:
            pass

        elif opCode == UartRxType.MESH_TRACKED_DEVICE_HEARTBEAT:
            pass


        ######################
        # Debug build events #
        ######################

        elif opCode == UartRxType.ADVERTISING_ENABLED:
            pass

        elif opCode == UartRxType.MESH_ENABLED:
            pass

        elif opCode == UartRxType.CROWNSTONE_ID:
            id = Conversion.int8_to_uint8(messagePacket.payload)
            UartEventBus.emit(DevTopics.ownCrownstoneId, id)

        elif opCode == UartRxType.MAC_ADDRESS:
            addr = Conversion.uint8_array_to_address(messagePacket.payload)
            if addr != "":
                UartEventBus.emit(DevTopics.ownMacAddress, addr)
            else:
                _LOGGER.warning("invalid address:", messagePacket.payload)

        elif opCode == UartRxType.ADC_CONFIG:
            # type is PowerCalculationsPacket
            parsedData = AdcConfigPacket(messagePacket.payload)
            UartEventBus.emit(DevTopics.newAdcConfigPacket, parsedData.getDict())

        elif opCode == UartRxType.ADC_RESTART:
            UartEventBus.emit(DevTopics.adcRestarted, None)

        elif opCode == UartRxType.POWER_LOG_CURRENT:
            # type is CurrentSamples
            parsedData = CurrentSamplesPacket(messagePacket.payload)
            UartEventBus.emit(DevTopics.newCurrentData, parsedData.getDict())

        elif opCode == UartRxType.POWER_LOG_VOLTAGE:
            # type is VoltageSamplesPacket
            parsedData = VoltageSamplesPacket(messagePacket.payload)
            UartEventBus.emit(DevTopics.newVoltageData, parsedData.getDict())

        elif opCode == UartRxType.POWER_LOG_FILTERED_CURRENT:
            # type is CurrentSamples
            parsedData = CurrentSamplesPacket(messagePacket.payload)
            UartEventBus.emit(DevTopics.newFilteredCurrentData, parsedData.getDict())

        elif opCode == UartRxType.POWER_LOG_FILTERED_VOLTAGE:
            # type is VoltageSamplesPacket
            parsedData = VoltageSamplesPacket(messagePacket.payload)
            UartEventBus.emit(DevTopics.newFilteredVoltageData, parsedData.getDict())

        elif opCode == UartRxType.POWER_LOG_POWER:
            # type is PowerCalculationsPacket
            parsedData = PowerCalculationPacket(messagePacket.payload)
            UartEventBus.emit(DevTopics.newCalculatedPowerData, parsedData.getDict())

        elif opCode == UartRxType.ASCII_LOG:
            timestamp = datetime.datetime.now()
            stringResult = ""
            for byte in messagePacket.payload:
                if byte < 128:
                    stringResult += chr(byte)
            logStr = f"ASCII LOG: [{timestamp.strftime(self.timestampFormat)}] {stringResult}"
            # sys.stdout.write(logStr)
            print(logStr.rstrip())

        elif opCode == UartRxType.FIRMWARESTATE:
            # no need to process this, that's in the test suite.
            pass

        else:
            _LOGGER.debug(f"Unknown opCode: {opCode}")

        parsedData = None
        



#!/usr/bin/env python3

"""An example that prints all Crownstone IDs seen on the mesh."""
import time

# Create new instance of Bluenet
from crownstone_uart import CrownstoneUart

uart = CrownstoneUart()

# Start up the USB bridge.
uart.initialize_usb_sync()
# you can alternatively do this async by
# await uart.initialize_usb()

# List the ids that have been seen
print("Listening for Crownstones on the mesh, this might take a while.")

# the try except part is just to catch a control+c, time.sleep does not appreciate being killed.
initial_time = time.time()
try:
	while uart.running:
		time.sleep(2)
		ids = uart.get_crownstone_ids()
		print("Crownstone IDs seen so far:", ids, "after", round(time.time() - initial_time), "seconds")
except KeyboardInterrupt:
	print("\nClosing example.... Thank you for your time!")
except:
	print("\nClosing example.... Thank you for your time!")

uart.stop()



class UartCommandHelloPacket:
    """
    UART command hello packet:
    1B flags
    """
    def __init__(self, encryptionRequired: bool = False, hasBeenSetUp: bool = False, hasInternet: bool = False, hasError: bool = False):
        self.encryptionRequired = encryptionRequired
        self.hasBeenSetUp = hasBeenSetUp
        self.hasInternet = hasInternet
        self.hasError = hasError

    def serialize(self):
        flags = 0 # Uint 8
        if self.encryptionRequired: flags = flags | (1 << 0)
        if self.hasBeenSetUp:       flags = flags | (1 << 1)
        if self.hasInternet:        flags = flags | (1 << 2)
        if self.hasError:           flags = flags | (1 << 3)
        return [flags]



# import signal  # used to catch control C
import logging
import queue
import time

from crownstone_core.protocol.BlePackets import ControlPacket
from crownstone_core.protocol.BluenetTypes import ControlType

from crownstone_uart.core.modules.AssetFilterHandler import AssetFilterHandler
from crownstone_uart.core.modules.ControlHandler import ControlHandler
from crownstone_uart.core.dataFlowManagers.UartWriter import UartWriter
from crownstone_uart.core.modules.MeshHandler import MeshHandler
from crownstone_uart.core.modules.MicroappHandler import MicroappHandler
from crownstone_uart.core.modules.SetupHandler import SetupHandler
from crownstone_uart.core.modules.StateHandler import StateHandler

from crownstone_uart.core.dataFlowManagers.StoneManager import StoneManager
from crownstone_uart.core.modules.UsbDevHandler import UsbDevHandler
import asyncio

from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.UartManager import UartManager
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket
from crownstone_uart.core.uart.UartTypes import UartTxType, UartMessageType
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket
from crownstone_uart.topics.SystemTopics import SystemTopics

_LOGGER = logging.getLogger(__name__)

class CrownstoneUart:
    __version__ = "2.7.0"

    def __init__(self):
        self.uartManager = None
        self.running = True

        self.manager_exception_queue = queue.Queue()
        self.uartManager = UartManager(self.manager_exception_queue)
        self.stoneManager = StoneManager()

        self.control = ControlHandler()
        self.state = StateHandler()
        self.mesh = MeshHandler(self.control)
        self.setup = SetupHandler(self.control)
        self.assetFilter = AssetFilterHandler(self.control)
        self.microapp = MicroappHandler(self.control, self.mesh)
        # only for development. Generally undocumented.
        self._usbDev = UsbDevHandler()

    def __del__(self):
        self.stop()

    def is_ready(self) -> bool:
        return self.uartManager.is_ready()

    async def initialize_usb(self, port = None, baudrate=230400, writeChunkMaxSize=64):
        """
        Initialize a Crownstone serial device.
            
        :param port: serial port of the USB. e.g. '/dev/ttyUSB0' or 'COM3'.
        :param baudrate: baudrate that should be used for this connection. default is 230400.
        :param writeChunkMaxSize: writing in chunks solves issues writing to certain JLink chips. A max chunkSize of 64 was found to work well for our case.
            For normal usage with Crownstones this is not required. a writeChunkMaxSize of 0 will not send the payload in chunks.
        
        This method is a coroutine.
        """
        self.uartManager.config(port, baudrate, writeChunkMaxSize)

        result = [False]
        def handleMessage(result, data):
            result[0] = True

        event = UartEventBus.subscribe(SystemTopics.connectionEstablished, lambda data: handleMessage(result, data))
        self.uartManager.start()

        while not result[0] and self.running:
            try:
                exception, exception_value, trace = self.manager_exception_queue.get(block=False)
                self.uartManager.join()
                self.stop()
                raise exception(exception_value)
            except queue.Empty:
                pass

            await asyncio.sleep(0.1)

        UartEventBus.unsubscribe(event)


    def initialize_usb_sync(self, port = None, baudrate=230400, writeChunkMaxSize=64):
        """
        Initialize a Crownstone serial device.
            
        :param port: serial port of the USB. e.g. '/dev/ttyUSB0' or 'COM3'.
        :param baudrate: baudrate that should be used for this connection. default is 230400.
        :param writeChunkMaxSize: writing in chunks solves issues writing to certain JLink chips. A max chunkSize of 64 was found to work well for our case.
            For normal usage with Crownstones this is not required. a writeChunkMaxSize of 0 will not send the payload in chunks.
        """
        self.uartManager.config(port, baudrate, writeChunkMaxSize)

        result = [False]

        def handleMessage(result, data):
            result[0] = True

        event = UartEventBus.subscribe(SystemTopics.connectionEstablished, lambda data: handleMessage(result, data))
        self.uartManager.start()

        try:
            while not result[0] and self.running:
                try:
                    exc = self.manager_exception_queue.get(block=False)
                    self.uartManager.join()
                    self.stop()
                    raise exc[0](exc[1])
                except queue.Empty:
                    pass

                time.sleep(0.1)

        except KeyboardInterrupt:
            print("\nClosing Crownstone Uart.... Thanks for your time!")
            self.stop()

        UartEventBus.unsubscribe(event)


    def stop(self):
        if self.uartManager is not None:
            self.uartManager.stop()
        self.running = False


    def switch_crownstone(self, crownstoneId: int, on: bool):
        """
        :param crownstoneId:
        :param on: Boolean
        :return:
        """
        if not on:
            self.mesh.turn_crownstone_off(crownstoneId)
        else:
            self.mesh.turn_crownstone_on(crownstoneId)


    def dim_crownstone(self, crownstoneId: int, switchVal: int):
        """
        :param crownstoneId:
        :param switchVal: 0% .. 100% or special values (SwitchValSpecial).
        :return:
        """

        self.mesh.set_crownstone_switch(crownstoneId, switchVal)


    def get_crownstone_ids(self):
        return self.stoneManager.getIds()

    def get_crownstones(self):
        return self.stoneManager.getStones()

    def uart_echo(self, payloadString):
        # wrap that in a control packet
        controlPacket = ControlPacket(ControlType.UART_MESSAGE).loadString(payloadString).serialize()

        # wrap that in a uart message
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()

        # finally, wrap it in an uart wrapper packet
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        # send over uart
        result = UartWriter(uartPacket).write_sync()



from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.Bitmasks import is_bit_set
from crownstone_core.util.BufferReader import BufferReader


class AssetIdReport(BasePacket):

    def __init__(self, data=None):
        self.assetId: int      = 0
        self.crownstoneId: int = 0
        self.rssi: int         = 0
        self.channel: int      = 0
        self.passedFilterIds   = []

        if data is not None:
            self.deserialize(data)

    def _deserialize(self, reader: BufferReader):
        self.assetId        = reader.getUInt8() + (reader.getUInt8() << 8) + (reader.getUInt8() << 16)
        self.crownstoneId   = reader.getUInt8()

        self.passedFilterIds.clear()
        passedFilterBitmask = reader.getUInt8()
        for i in range(0, 8):
            if is_bit_set(passedFilterBitmask, i):
                self.passedFilterIds.append(i)
        
        self.rssi           = reader.getInt8()
        self.channel        = reader.getUInt8()

    def __str__(self):
        return f"AssetIdReport(" \
               f"assetId={self.assetId} " \
               f"crownstoneId={self.crownstoneId} " \
               f"rssi={self.rssi} " \
               f"channel={self.channel} " \
               f"passedFilterIds={self.passedFilterIds})"



from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.dataFlowManagers.StoneStateManager import StoneStateManager
from crownstone_uart.topics.SystemTopics import SystemTopics


class StoneManager:
    
    def __init__(self):
        self.stones = {}
        self.stateManager = StoneStateManager()
        UartEventBus.subscribe(SystemTopics.newCrownstoneFound, self.handleNewStoneFromScan)

    def getIds(self):
        ids = []
        for stoneId, data in self.stones.items():
            ids.append(stoneId)

        return ids
    
    
    def getStones(self):
        return self.stones
    
    
    def handleNewStoneFromScan(self, stoneId):
        if stoneId in self.stones:
            self.stones[stoneId]["available"] = True
        else:
            self.stones[stoneId] = {"available": True, "id": stoneId}


import logging
from typing import List

from crownstone_core.packets.assetFilter.FilterCommandPackets import FilterSummariesPacket, FilterSummaryPacket
from crownstone_core.packets.assetFilter.builders.AssetFilter import AssetFilter
from crownstone_core.packets.assetFilter.util import AssetFilterMasterCrc
from crownstone_core.packets.assetFilter.util.AssetFilterChunker import FilterChunker
from crownstone_core.packets.assetFilter.util.AssetFilterSyncer import AssetFilterSyncer
from crownstone_core.protocol.ControlPackets import ControlPacketsGenerator

from crownstone_uart.core.modules.ControlHandler import ControlHandler

_LOGGER = logging.getLogger(__name__)

class AssetFilterHandler:
    def __init__(self, control: ControlHandler):
        self.control = control
        pass

    async def setFilters(self, filters: List[AssetFilter], masterVersion: int = None) -> int:
        """
        Makes sure the given filters are set at the Crownstone.
        Uploads and removes filters where necessary.
        :param filters:           The asset filter to be uploaded.
        :param masterVersion:     The new master version. If None, the master version will be increased by 1.
        :return:                  The new master version.
        """
        _LOGGER.info(f"setFilters")
        summaries = await self.getFilterSummaries()
        syncer = AssetFilterSyncer(summaries, filters, masterVersion)
        if not syncer.commitRequired:
            return syncer.masterVersion

        for filterId in syncer.removeIds:
            await self.removeFilter(filterId)

        for filter in filters:
            if filter.getFilterId() in syncer.uploadIds:
                await self.uploadFilter(filter)

        await self.commitFilterChanges(syncer.masterVersion, filters)
        return syncer.masterVersion

    async def getFilterSummaries(self) -> FilterSummariesPacket:
        """
        Get a summary of the filters that are on the Crownstones.
        This can be used to determine:
        - Which filters should be changed.
        - What the next master version should be.
        - How much space there is left for new filters.
        - The new master CRC.

        :return:   The filter summaries packet.
        """
        _LOGGER.info(f"getFilterSummaries")
        result = await self.control._writeControlAndGetResult(ControlPacketsGenerator.getGetFilterSummariesPacket())
        summaries = FilterSummariesPacket(result.payload)
        return summaries

    async def uploadFilter(self, filter: AssetFilter):
        """
        Upload an asset filter to the Crownstones.
        Once all changes are made, don't forget to commit them.

        :param filter:  The asset filter to be uploaded.
        """
        _LOGGER.info(f"uploadFilter {filter}")
        chunker = FilterChunker(filter, 128)
        for i in range(0, chunker.getAmountOfChunks()):
            chunk = chunker.getChunk()
            await self.control._writeControlAndWaitForSuccess(ControlPacketsGenerator.getUploadFilterPacket(chunk))

    async def removeFilter(self, filterId):
        """
        Remove an asset filter from the Crownstones.
        Once all changes are made, don't forget to commit them.

        :param filterId:     The filter ID to be removed.
        """
        _LOGGER.info(f"removeFilter id={filterId}")
        return await self.control._writeControlAndWaitForSuccess(ControlPacketsGenerator.getRemoveFilterPacket(filterId))

    async def commitFilterChanges(self, masterVersion: int, filters: List[AssetFilter], filterSummaries: List[FilterSummaryPacket] = None):
        """
        Commit the changes made by upload and/or remove.

        :param masterVersion:     The new master version, should be higher than previous master version.
        :param filters:           A list of asset filters with filter ID, that are uploaded to the Crowstone.
        :param filterSummaries :  A list of filter summaries that are already on the Crownstone.
        """
        _LOGGER.info(f"commitFilterChanges masterVersion={masterVersion}")
        masterCrc = AssetFilterMasterCrc.get_master_crc_from_filters(filters, filterSummaries)
        await self.control._writeControlAndWaitForSuccess(ControlPacketsGenerator.getCommitFilterChangesPacket(masterVersion, masterCrc))



from crownstone_core.Exceptions import CrownstoneException, CrownstoneError
from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader
from typing import List

class RssiBetweenStonesPacket(BasePacket):

    def __init__(self, data=None):

        self.type: int                 = 0

        # Stone ID of the stone that received a message.
        self.receiverId: int           = 0

        # Stone ID of the stone that sent a message.
        self.senderId: int             = 0

        # RSSI between sender and receiver at channel 37, 38, 39.
        # Will be None if there is no data for that channel.
        self.rssiPerChannel: List[int or None] = [None, None, None]

        # How many seconds ago the sender was last seen by the receiver.
        self.lastSeenSecondsAgo: int   = 0

        # Number that is increased by 1 each time the receiver sends this report.
        self.reportNumber: int         = 0

        if data is not None:
            self.deserialize(data)

    def _deserialize(self, reader: BufferReader):
        self.type               = reader.getUInt8()
        if self.type != 0:
            raise CrownstoneException(CrownstoneError.UNKNOWN_TYPE, f"Unknown type {self.type}")

        self.receiverId         = reader.getUInt8()
        self.senderId           = reader.getUInt8()
        for i in range(0, 3):
            rssi = reader.getInt8()
            if rssi == 0:
                rssi = None
            self.rssiPerChannel[i] = rssi
        self.lastSeenSecondsAgo = reader.getUInt8()
        self.reportNumber       = reader.getUInt8()

    def __str__(self):
        return f"RssiBetweenStonesPacket(" \
               f"receiverId={self.receiverId} " \
               f"senderId={self.senderId} "\
               f"rssi=[{self.rssiPerChannel[0]}, {self.rssiPerChannel[1]}, {self.rssiPerChannel[2]}] " \
               f"lastSeenSecondsAgo={self.lastSeenSecondsAgo} " \
               f"reportNumber={self.reportNumber})"



class UartUtil:

	UART_START_CHAR =       0x7E
	UART_ESCAPE_CHAR =      0x5C
	UART_ESCAPE_FLIP_MASK = 0x40

	@staticmethod
	def uartEscape(val):
		if isinstance(val, list):
			# Escape special chars:
			escapedMsg = []
			for c in val:
				if c == UartUtil.UART_ESCAPE_CHAR or c == UartUtil.UART_START_CHAR:
					escapedMsg.append(UartUtil.UART_ESCAPE_CHAR)
					c = UartUtil.uartEscape(c)
				escapedMsg.append(c)
			return escapedMsg
		else:
			return val ^ UartUtil.UART_ESCAPE_FLIP_MASK

	@staticmethod
	def uartUnescape(val):
		return val ^ UartUtil.UART_ESCAPE_FLIP_MASK



import logging

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader

_LOGGER = logging.getLogger(__name__)

class UartLogHeaderPacket(BasePacket):
	"""
	UART log header packet:
	4B filename hash
	2B line number
	1B log level
	1B flags:
	- bit 0 = newLine
	- bit 1 = reverse
	"""

	def __init__(self, data = None):
		self.fileNameHash = None
		self.lineNr = None
		self.logLevel = None
		# Flags:
		self.newLine = False
		self.reverse = False

		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		self.fileNameHash = reader.getUInt32()
		self.lineNr = reader.getUInt16()
		self.logLevel = reader.getUInt8()

		flags = reader.getUInt8()
		self.newLine = (flags & (1 << 0)) != 0
		self.reverse = (flags & (1 << 1)) != 0

	def __str__(self):
		return f"UartLogHeaderPacket(" \
		       f"fileNameHash={self.fileNameHash}, " \
		       f"lineNr={self.lineNr}, " \
		       f"logLevel={self.logLevel}, " \
		       f"newLine={self.newLine}, " \
		       f"reverse={self.reverse})"



from crownstone_core.util.BufferReader import BufferReader


class UartCrownstoneStatusPacket:
    """
    UART crownstone status packet:
    1B flags
    """

    def __init__(self, buffer: BufferReader or list or bytearray):
        if isinstance(buffer, BufferReader):
            reader = buffer
        else:
            reader = BufferReader(buffer)

        self.flags = reader.getUInt8()

        # Parse flags
        self.encryptionRequired = self.flags & (1 << 0) != 0
        self.hasBeenSetUp       = self.flags & (1 << 1) != 0
        self.hubMode            = self.flags & (1 << 2) != 0
        self.hasError           = self.flags & (1 << 3) != 0



from crownstone_core import Conversion
from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader


class AssetMacReport(BasePacket):

    def __init__(self, data=None):
        self.assetMacAddress: str = ""
        self.crownstoneId: int    = 0
        self.rssi: int            = 0
        self.channel: int         = 0

        if data is not None:
            self.deserialize(data)

    def _deserialize(self, reader: BufferReader):
        self.assetMacAddress = Conversion.uint8_array_to_address(bytearray(reader.getBytes(6)))
        self.crownstoneId    = reader.getUInt8()
        self.rssi            = reader.getInt8()
        self.channel         = reader.getUInt8()

    def __str__(self):
        return f"AssetMacReport(" \
               f"assetMacAddress={self.assetMacAddress} " \
               f"crownstoneId={self.crownstoneId} " \
               f"rssi={self.rssi} " \
               f"channel={self.channel})"



import logging
import threading
import time
import sys
import queue

from crownstone_core.protocol.BlePackets import ControlPacket
from crownstone_core.protocol.BluenetTypes import ControlType

from crownstone_uart.core.dataFlowManagers.Collector import Collector
from crownstone_uart.core.uart.uartPackets.UartCommandHelloPacket import UartCommandHelloPacket
from crownstone_uart.core.uart.uartPackets.UartCrownstoneHelloPacket import UartCrownstoneHelloPacket
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket

from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket
from crownstone_uart.core.uart.UartTypes import UartTxType, UartMessageType
from crownstone_uart.core.uart.UartBridge import UartBridge

from crownstone_uart.topics.UartTopics import UartTopics
from crownstone_uart.topics.SystemTopics import SystemTopics

from crownstone_uart.Exceptions import UartException

from serial.tools import list_ports

_LOGGER = logging.getLogger(__name__)

class UartManager(threading.Thread):

    def __init__(self, exception_queue):
        self.port = None     # Port configured by user.
        self.baudRate = 230400
        self.writeChunkMaxSize = 0
        self.manager_exception_queue = exception_queue
        self.running = True
        self._availablePorts = list(list_ports.comports())
        self._attemptingIndex = 0
        self._uartBridge = None
        self.ready = False

        self.eventId = UartEventBus.subscribe(SystemTopics.connectionClosed, self.resetEvent)

        self.custom_port_set = False

        threading.Thread.__init__(self)

    def __del__(self):
        self.stop()

    def config(self, port, baudRate = 230400, writeChunkMaxSize=0):
        if port is not None:
            self.custom_port_set = True
        self.port            = port
        self.baudRate        = baudRate
        self.writeChunkMaxSize = writeChunkMaxSize

    def run(self):
        try:
            self.initialize()
        except (UartException, BaseException):
            self.manager_exception_queue.put(sys.exc_info())

    def stop(self):
        self.running = False
        UartEventBus.unsubscribe(self.eventId)
        if self._uartBridge is not None:
            self._uartBridge.stop()

    def resetEvent(self, eventData=None):
        if self.ready:
            self._uartBridge.stop()
            while self._uartBridge.running:
                time.sleep(0.1)
            self.reset()

            self.ready = False
            self.initialize()

    def reset(self):
        if self.running:
            self._attemptingIndex = 0
            self._availablePorts = list(list_ports.comports())
            self._uartBridge = None
            self.port = None

    def echo(self, string):
        controlPacket = ControlPacket(ControlType.UART_MESSAGE).loadString(string).serialize()
        uartMessage   = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket    = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

    def writeHello(self):
        helloPacket = UartCommandHelloPacket().serialize()
        uartMessage = UartMessagePacket(UartTxType.HELLO, helloPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

    def initialize(self):
        if not self.custom_port_set:
            _LOGGER.warning(F"By not providing a specific port to find the Crownstone dongle, we will try to connect and handshake with all available ports one by one until we find the dongle."
                            F"\nPorts that will be checked are:"
                            F"\n{[port.device for port in self._availablePorts]}")

        while self.running and not self.ready:
            self.ready = False
            _LOGGER.debug(F"Initializing... {self.port} {self.baudRate}")

            # if the user provides his own port, we check if it is in the list of ports we have
            # if it exists, we start there and skip the handshake.
            # if it is not in the list, we attempt the connection regardless, this might change in the future.
            if self.port is not None:
                found_port = False
                index = 0
                # check if the provided port is among the listed ports.
                for testPort in self._availablePorts:
                    if self.port == testPort.device:
                        found_port = True
                        self._attemptingIndex = index
                        break
                    index += 1
                # if it is not in the list of available ports
                if not found_port:
                    # we will attempt it anyway
                    _LOGGER.warning(F"Could not find provided port {self.port}")
                else:
                    _LOGGER.debug(F"Setup connection to {self.port}")
                self.setupConnection(self.port, performHandshake=False)
                continue


            if self.port is None:
                if self._attemptingIndex >= len(self._availablePorts): # this also catches len(self._availablePorts) == 0
                    _LOGGER.warning("No Crownstone USB connected? Retrying...")
                    time.sleep(1)
                    self.reset()
                else:
                    self._attemptConnection(self._attemptingIndex)


    def _attemptConnection(self, index, handshake=True):
        attemptingPort = self._availablePorts[index]
        self.setupConnection(attemptingPort.device, handshake)


    def setupConnection(self, port, performHandshake=True):
        _LOGGER.debug(F"Setting up connection... port={port} baudRate={self.baudRate} performHandshake={performHandshake}")
        bridge_exception_queue = queue.Queue()
        self._uartBridge = UartBridge(bridge_exception_queue, port, self.baudRate, self.writeChunkMaxSize)
        self._uartBridge.start()

        def initialize_bridge():
            # handle exceptions that happen in thread while initializing
            while not self._uartBridge.started and self.running:
                try:
                    exception, exception_value, trace = bridge_exception_queue.get(block=False)
                    self._uartBridge.join()
                    raise exception(exception_value)
                except queue.Empty:
                    pass
                    
                time.sleep(0.1)
        
        # attempt serial connection with the dongle
        initialize_bridge()

        # this is true by default if no handshake is required
        handshake_succesfull = True

        if performHandshake:
            collector = Collector(timeout=0.25, topic=UartTopics.hello)
            self.writeHello()
            reply = collector.receive_sync()

            handshake_succesfull = False
            if isinstance(reply, UartCrownstoneHelloPacket):
                handshake_succesfull = True
                _LOGGER.debug("Handshake successful")
            else:
                if self.port == port:
                    _LOGGER.warning("Handshake failed: no reply from the crownstone.")
                else:
                    _LOGGER.debug("Handshake failed: no reply from the crownstone.")

        if not handshake_succesfull:
            _LOGGER.debug("Reinitialization required")
            self._attemptingIndex += 1
            self._uartBridge.stop()

            initialize_bridge()

        else:
            _LOGGER.info("Connection established to {}".format(port))
            self.port = port
            self.ready = True
            UartEventBus.emit(SystemTopics.connectionEstablished)



    def is_ready(self) -> bool:
        return self.ready



from enum import Enum


class UartError(Enum):
    """Error types for Uart exceptions."""
    NO_CROWNSTONE_UART_DEVICE_AVAILABLE = "NO_CROWNSTONE_UART_DEVICE_AVAILABLE"


class UartBridgeError(Enum):
    """Error types for Uart Bridge exceptions."""
    CANNOT_OPEN_SERIAL_CONTROLLER       = "CANNOT_OPEN_SERIAL_CONTROLLER"

    def __str__(self) -> str:
        """Return value of the error."""
        return self.value


class UartException(Exception):
    """Raised on errors in the Uart Manager."""



import logging

from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.packets.microapp.MicroappInfoPacket import MicroappInfoPacket
from crownstone_core.packets.microapp.MicroappMessagePacket import MicroappMessagePacket
from crownstone_core.packets.microapp.MicroappUploadPacket import MicroappUploadPacket
from crownstone_core.protocol.BlePackets import ControlPacket
from crownstone_core.protocol.BluenetTypes import ControlType
from crownstone_uart.core.modules.MeshHandler import MeshHandler
from crownstone_uart.core.modules.ControlHandler import ControlHandler

_LOGGER = logging.getLogger(__name__)

class MicroappHandler:
    def __init__(self, control: ControlHandler, mesh: MeshHandler):
        self._control = control
        self._mesh = mesh
        pass

    # The UART RX buffer of the firmware is only 192B.
    MAX_CHUNK_SIZE = 128

    async def getMicroappInfo(self) -> MicroappInfoPacket:
        resultPacket = await self._control._writeControlAndGetResult(ControlPacket(ControlType.MICROAPP_GET_INFO).serialize())
        _LOGGER.info(f"getMicroappInfo {resultPacket}")
        infoPacket = MicroappInfoPacket(resultPacket.payload)
        return infoPacket

    async def uploadMicroapp(self, data: bytearray, index: int = 0, protocol: int = 0, chunkSize: int = MAX_CHUNK_SIZE):
        chunkSize = min(chunkSize, MicroappHandler.MAX_CHUNK_SIZE)

        for i in range(0, len(data), chunkSize):
            chunk = data[i : i + chunkSize]
            # Pad the chunk with 0xFF, so the size is a multiple of 4.
            if len(chunk) % 4:
                if isinstance(chunk, bytes):
                    chunk = bytearray(chunk)
                chunk.extend((4 - (len(chunk) % 4)) * [0xFF])
            await self._uploadMicroappChunk(index, protocol, chunk, i)

    async def _uploadMicroappChunk(self, index: int, protocol: int, data: bytearray, offset: int):
        _LOGGER.info(f"Upload microapp chunk index={index} offset={offset} size={len(data)}")
        header = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        packet = MicroappUploadPacket(header, offset, data)
        controlPacket = ControlPacket(ControlType.MICROAPP_UPLOAD).loadByteArray(packet.serialize()).serialize()
        await self._sendCommand(controlPacket)
        _LOGGER.info(f"uploaded chunk offset={offset}")

    async def validateMicroapp(self, index: int, protocol: int, crownstoneId: int = None):
        # Validate an uploaded microapp.
        # When the crownstone ID is given, the command will be sent via the mesh.
        packet = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_VALIDATE).loadByteArray(packet.serialize()).serialize()
        await self._sendCommand(controlPacket, crownstoneId)

    async def enableMicroapp(self, index: int, protocol: int, crownstoneId: int = None):
        # Enable a validated microapp.
        # When the crownstone ID is given, the command will be sent via the mesh.
        packet = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_ENABLE).loadByteArray(packet.serialize()).serialize()
        await self._sendCommand(controlPacket, crownstoneId)

    async def disableMicroapp(self, index: int, protocol: int, crownstoneId: int = None):
        # Disable a microapp.
        # When the crownstone ID is given, the command will be sent via the mesh.
        packet = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_DISABLE).loadByteArray(packet.serialize()).serialize()
        await self._sendCommand(controlPacket, crownstoneId)


    async def removeMicroapp(self, index: int, protocol: int, crownstoneId: int = None):
        # Remove a microapp.
        # When the crownstone ID is given, the command will be sent via the mesh.
        packet = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_REMOVE).loadByteArray(packet.serialize()).serialize()
        await self._sendCommand(controlPacket, crownstoneId)

    async def sendMessage(self, index: int, protocol: int, data: bytearray):
        _LOGGER.info(f"Send message to microapp index={index}")
        header = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        packet = MicroappMessagePacket(header, data)
        controlPacket = ControlPacket(ControlType.MICROAPP_MESSAGE).loadByteArray(packet.serialize()).serialize()
        await self._sendCommand(controlPacket)

    async def _sendCommand(self, controlPacket, crownstoneId: int = None):
        if crownstoneId is None:
            await self._control._writeControlAndWaitForSuccess(controlPacket)
        else:
            meshResult = await self._mesh._command_via_mesh_broadcast_acked([crownstoneId], controlPacket)
            if not meshResult.success:
                _LOGGER.warning(f"Microapp command via mesh: ack={meshResult.resultCode} success={meshResult.success}")



import time, math, asyncio
from typing import List

from crownstone_core import Conversion
from crownstone_core.Exceptions import CrownstoneException
from crownstone_core.protocol.BlePackets import ControlPacket, ControlStateSetPacket
from crownstone_core.protocol.BluenetTypes import ControlType, StateType, ResultValue, SwitchValSpecial
from crownstone_core.protocol.ControlPackets import ControlPacketsGenerator
from crownstone_core.protocol.MeshPackets import MeshMultiSwitchPacket, StoneMultiSwitchPacket, MeshSetStatePacket, MeshBroadcastPacket, MeshBroadcastAckedPacket
from crownstone_core.util.Timestamp import getCorrectedLocalTimestamp

from crownstone_uart.core.containerClasses.MeshResult import MeshResult
from crownstone_uart.core.dataFlowManagers.BatchCollector import BatchCollector
from crownstone_uart.core.dataFlowManagers.Collector import Collector
from crownstone_uart.core.UartEventBus import UartEventBus
from crownstone_uart.core.modules.ControlHandler import ControlHandler
from crownstone_uart.core.uart.uartPackets.UartMessagePacket import UartMessagePacket
from crownstone_uart.core.uart.UartTypes import UartTxType, UartMessageType
from crownstone_uart.core.uart.uartPackets.UartWrapperPacket import UartWrapperPacket
from crownstone_uart.topics.SystemTopics import SystemTopics


class MeshHandler:
    def __init__(self, control: ControlHandler):
        self.control = control
        pass

    # TODO: make async, wait for uart reply.
    def turn_crownstone_on(self, crownstone_id: int):
        self._switch_crownstone(crownstone_id, SwitchValSpecial.SMART_ON)

    # TODO: make async, wait for uart reply.
    def turn_crownstone_off(self, crownstone_id: int):
        self._switch_crownstone(crownstone_id, 0)

    # TODO: make async, wait for uart reply.
    def set_crownstone_switch(self, crownstone_id: int, switch_val: int):
        """
        :param crownstone_id:
        :param switch_val: 0% .. 100% or special value (SwitchValSpecial).
        :return:
        """

        self._switch_crownstone(crownstone_id, switch_val)

    # TODO: make async, wait for uart reply.
    def _switch_crownstone(self,crownstone_id: int, switch_val: int):
        """
        :param crownstone_id:
        :param switch_val: 0% .. 100% or special value (SwitchValSpecial).
        :return:
        """

        # create a stone switch state packet to go into the multi switch
        stoneSwitchPacket = StoneMultiSwitchPacket(crownstone_id, switch_val)

        # wrap it in a mesh multi switch packet
        meshMultiSwitchPacket = MeshMultiSwitchPacket([stoneSwitchPacket]).serialize()

        # wrap that in a control packet
        controlPacket = ControlPacket(ControlType.MULTISWITCH).loadByteArray(meshMultiSwitchPacket).serialize()

        # wrap that in a uart message
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()

        # finally wrap it in a uart wrapper packet
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        # send over uart
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)


    async def set_time(self, timestamp = None):
        if timestamp is None:
            timestamp = math.ceil(time.time())

        localizedTimeStamp = getCorrectedLocalTimestamp(timestamp)

        time_packet = ControlPacketsGenerator.getSetTimePacket(localizedTimeStamp)
        await self._command_via_mesh_broadcast(time_packet)


    async def send_no_op(self):
        control_packet = ControlPacket(ControlType.NO_OPERATION)
        await self._command_via_mesh_broadcast(control_packet.serialize())

    async def reset_rssi_between_stones(self, crownstone_uid_array: List[int] = None):
        control_packet = ControlPacket(ControlType.RESET_RSSI_BETWEEN_STONES)
        if (crownstone_uid_array == None):
            return await self._command_via_mesh_broadcast(control_packet.serialize())
        else:
            return await self._command_via_mesh_broadcast_acked(crownstone_uid_array, control_packet.serialize())

    async def reset_errors(self):
        control_packet = ControlPacketsGenerator.getResetErrorPacket(0xFFFFFFFF)
        await self._command_via_mesh_broadcast(control_packet)

    async def set_tx_power(self, crownstone_uid_array: List[int], txPower: int):
        statePacket = ControlStateSetPacket(StateType.TX_POWER)
        statePacket.loadInt8(txPower)
        return await self._command_via_mesh_broadcast_acked(crownstone_uid_array, statePacket.serialize())

    async def set_ibeacon_uuid(self, crownstone_id: int, uuid: str, index: int = 0) -> MeshResult:
        """
        :param crownstone_id: int crownstoneUid, 1-255
        :param uuid:  string: "d8b094e7-569c-4bc6-8637-e11ce4221c18"
        :param index: for the normal uuid, index = 0, when alternating you also need to define 1 in a
                      followup command. Usually 0 has already been set by the setup procedure.
        :return:
        """
        statePacket = ControlStateSetPacket(StateType.IBEACON_UUID, index)
        statePacket.loadByteArray(Conversion.ibeaconUUIDString_to_reversed_uint8_array(uuid))
        return await self._set_state_via_mesh_acked(crownstone_id, statePacket.serialize())


    async def set_ibeacon_major(self, crownstone_id: int, major: int, index: int = 0) -> MeshResult:
        """
        :param crownstone_id: int crownstoneUid, 1-255
        :param major:  int: uint16 0-65535
        :param index: for the normal uuid, index = 0, when alternating you also need to define 1 in a
                      followup command. Usually 0 has already been set by the setup procedure.
        :return:
        """
        statePacket = ControlStateSetPacket(StateType.IBEACON_MAJOR, index)
        statePacket.loadUInt16(major)
        return await self._set_state_via_mesh_acked(crownstone_id, statePacket.serialize())


    async def set_ibeacon_minor(self, crownstone_id: int, minor: int, index: int = 0) -> MeshResult:
        """
        :param crownstone_id: int crownstoneUid, 1-255
        :param minor:  int: uint16 0-65535
        :param index: for the normal uuid, index = 0, when alternating you also need to define 1 in a
                      followup command. Usually 0 has already been set by the setup procedure.
        :return:
        """
        statePacket = ControlStateSetPacket(StateType.IBEACON_MINOR, index)
        statePacket.loadUInt16(minor)
        return await self._set_state_via_mesh_acked(crownstone_id, statePacket.serialize())


    async def periodically_activate_ibeacon_index(self, crownstone_uid_array: List[int], index : int, interval_seconds: int, offset_seconds: int = 0) -> MeshResult:
        """
        You need to have 2 stored ibeacon payloads (state index 0 and 1) in order for this to work. This can be done by the set_ibeacon methods
        available in this class.

        Once the interval starts, it will set this ibeacon ID to be active. In order to have 2 ibeacon payloads interleaving, you have to call this method twice.
        To interleave every minute
        First,    periodically_activate_ibeacon_index, index 0, interval = 120 (2 minutes), offset = 0
        Secondly, periodically_activate_ibeacon_index, index 1, interval = 120 (2 minutes), offset = 60

        This will change the active ibeacon payload every minute:
        T        = 0.............60.............120.............180.............240
        activeId = 0.............1...............0...............1...............0
        period_0 = |------------120s-------------|--------------120s-------------|
        :param crownstone_uid_array:
        :param index:
        :param interval_seconds:
        :param offset_seconds:
        :return:
        """

        ibeaconConfigPacket = ControlPacketsGenerator.getIBeaconConfigIdPacket(index, offset_seconds, interval_seconds)
        return await self._command_via_mesh_broadcast_acked(crownstone_uid_array, ibeaconConfigPacket)


    async def stop_ibeacon_interval_and_set_index(self, crownstone_uid_array: List[int], index) -> MeshResult:
        """
        This method stops the interleaving for the specified ibeacon payload at that index.
        :param crownstone_uid_array:
        :param index:
        :return:
        """
        indexToEndWith = index
        indexToStartWith = 0
        if index == 0:
            indexToStartWith = 1

        ibeaconConfigPacketStart  = ControlPacketsGenerator.getIBeaconConfigIdPacket(indexToStartWith, 0, 0)
        ibeaconConfigPacketFinish = ControlPacketsGenerator.getIBeaconConfigIdPacket(indexToEndWith,   0, 0)

        meshResult = MeshResult(crownstone_uid_array)

        initialResult = await self._command_via_mesh_broadcast_acked(crownstone_uid_array, ibeaconConfigPacketStart)

        meshResult.merge(initialResult)
        successfulIds = meshResult.get_successful_ids()
        if len(successfulIds) == 0:
            return meshResult

        secondResult = await self._command_via_mesh_broadcast_acked(successfulIds, ibeaconConfigPacketFinish)

        # if we succeeded in the initial phase, we should be able to finish the second case.
        failed_second_part = meshResult.compare_get_failed(secondResult)
        iterations = 0
        while len(failed_second_part) > 0 and iterations < 5:
            secondResult = await self._command_via_mesh_broadcast_acked(failed_second_part, ibeaconConfigPacketFinish)
            failed_second_part = meshResult.compare_get_failed(secondResult)
            iterations += 1

        meshResult.merge(secondResult)
        meshResult.conclude()

        return meshResult


    async def _set_state_via_mesh_acked(self, crownstone_id: int, packet: bytearray) -> MeshResult:
        # 1:1 message to N crownstones with acks (only N = 1 supported for now)
        # flag value: 2
        corePacket    = MeshSetStatePacket(crownstone_id, packet).serialize()
        controlPacket = ControlPacket(ControlType.MESH_COMMAND).loadByteArray(corePacket).serialize()
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        resultCollector     = Collector(timeout=2,  topic=SystemTopics.resultPacket)
        individualCollector = BatchCollector(timeout=15, topic=SystemTopics.meshResultPacket)
        finalCollector      = Collector(timeout=15, topic=SystemTopics.meshResultFinalPacket)

        # send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        # wait for the collectors to fill
        commandResultData = await resultCollector.receive()

        if commandResultData is not None:
            if commandResultData.resultCode is ResultValue.BUSY:
                await asyncio.sleep(0.2)
                return await self._set_state_via_mesh_acked(crownstone_id, packet)
            elif commandResultData.resultCode is not ResultValue.SUCCESS:
                raise CrownstoneException(commandResultData.resultCode, "Command has failed.")

        return await self._handleCollectors([crownstone_id], individualCollector, finalCollector)




    async def _command_via_mesh_broadcast(self, packet: bytearray):
        # Send a control command via the mesh to all crownstones.
        # Not all control commands are allowed.
        # broadcast to all:
        # value: 1
        corePacket = MeshBroadcastPacket(packet).serialize()
        controlPacket = ControlPacket(ControlType.MESH_COMMAND).loadByteArray(corePacket).serialize()
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        resultCollector = Collector(timeout=2, topic=SystemTopics.resultPacket)

        # send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        # wait for the collectors to fill
        commandResultData = await resultCollector.receive()

        if commandResultData is not None:
            if commandResultData.resultCode is ResultValue.BUSY:
                await asyncio.sleep(0.2)
                return await self._command_via_mesh_broadcast(packet)
            elif commandResultData.resultCode is not ResultValue.SUCCESS:
                raise CrownstoneException(commandResultData.resultCode, "Command has failed.")

        await asyncio.sleep(0.1)


    async def _command_via_mesh_broadcast_acked(self, crownstone_uid_array: List[int], packet: bytearray) -> MeshResult:
        # Send a control command via the mesh to a list of crownstones.
        # Not all control commands are allowed.
        # If the list of crownstone IDs is larger than 1, the command will be broadcasted to all,
        # but retried until all IDs in list have acked or timed out.
        # Otherwise, the command will only be received by the single crownstone ID.
        corePacket    = MeshBroadcastAckedPacket(crownstone_uid_array, packet).serialize()
        controlPacket = ControlPacket(ControlType.MESH_COMMAND).loadByteArray(corePacket).serialize()
        uartMessage = UartMessagePacket(UartTxType.CONTROL, controlPacket).serialize()
        uartPacket = UartWrapperPacket(UartMessageType.UART_MESSAGE, uartMessage).serialize()

        resultCollector     = Collector(timeout=2, topic=SystemTopics.resultPacket)
        individualCollector = BatchCollector(timeout=15, topic=SystemTopics.meshResultPacket)
        finalCollector      = Collector(timeout=15, topic=SystemTopics.meshResultFinalPacket)

        # send the message to the Crownstone
        UartEventBus.emit(SystemTopics.uartWriteData, uartPacket)

        # wait for the collectors to fill
        commandResultData = await resultCollector.receive()
        if commandResultData is not None:
            if commandResultData.resultCode is ResultValue.BUSY:
                await asyncio.sleep(0.2)
                return await self._command_via_mesh_broadcast_acked(crownstone_uid_array, packet)
            elif commandResultData.resultCode is not ResultValue.SUCCESS:
                raise CrownstoneException(commandResultData.resultCode, "Command has failed.")

        return await self._handleCollectors(crownstone_uid_array, individualCollector, finalCollector)



    async def _handleCollectors(self, crownstone_uid_array: List[int], individualCollector: BatchCollector, finalCollector: Collector) -> MeshResult:
        meshResult = MeshResult(crownstone_uid_array)

        # await the amount of times we have ID's to deliver the message to
        for uid in crownstone_uid_array:
            # Since the individualCollector is using topic meshResultPacket, the data is [Crownstone ID, ResultPacket]
            individualData = await individualCollector.receive()
            if individualData is not None:
                meshResult.collect_ack(individualData[0], individualData[1].resultCode == ResultValue.SUCCESS)
            individualCollector.clear()

        individualCollector.cleanup()

        finalData = await finalCollector.receive()
        if finalData is not None:
            meshResult.resultCode = finalData.resultCode

        meshResult.conclude()

        return meshResult



#!/usr/bin/env python3

"""An example that prints all Crownstone IDs seen on the mesh."""

import time, json

from crownstone_uart import CrownstoneUart, UartEventBus, UartTopics

uart = CrownstoneUart()

# Start up the USB bridge.
uart.initialize_usb_sync()

def showNewData(data):
	print("New data received!")
	print(data)
	print("-------------------")

	print("PING!")
	uart.uart_echo("PONG!")

def showUartMessage(data):
	print("Received Uart Message " + data["string"])

# Set up event listeners
UartEventBus.subscribe(UartTopics.newDataAvailable, showNewData)
UartEventBus.subscribe(UartTopics.uartMessage, showUartMessage)

print("Listening for Crownstones on the mesh, this might take a while.")
# the try except part is just to catch a control+c, time.sleep does not appreciate being killed.
try:
	while uart.running:
		time.sleep(1)
except KeyboardInterrupt:
	print("Closing example.... Thanks for your time!")

uart.stop()
